import torch
from torch.nn.modules.loss import _Loss


class CrossEntropyLoss(_Loss):
    def __init__(self):
        super(CrossEntropyLoss, self).__init__()
        self.loss_func = torch.nn.CrossEntropyLoss(reduction="sum", ignore_index=0)

    def forward(self, logits, labels, mask=None):
        loss = torch.zeros(1, device=logits.device)
        batch_size = logits.size(0)

        for logit, label, m in zip(logits, labels, mask):
            loss += self.loss_func(logit[m], label[m])

        return torch.sum(loss / batch_size)



import json
import random
import warnings
from logging import getLogger

import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import numpy as np
import torch
from marine.data.feature.feature_table import RAW_FEATURE_KEYS
from marine.utils.g2p_util import mora2phon, pron2mora
from marine.utils.regex import has_longvowel
from sklearn.model_selection import train_test_split

logger = getLogger(__name__)

LABEL_TABLE = {
    "intonation_phrase_boundary": {
        "O": 1,
        "B": 2,
    }
}
IGNORE_TARGET_MORA_FOR_ACCENT = ["ー", "ッ"]
JSON_LOG_STATUS_TO_TAG = {True: "success", False: "fail"}
JSON_LOG_TAG_TO_STATUS = {tag: status for status, tag in JSON_LOG_STATUS_TO_TAG.items()}

OPEN_JTALK_FEATURE_INDEX_TABLE = {
    "surface": 0,
    "pos": [1, 2, 3, 4],
    "c_type": 5,
    "c_form": 6,
    "pron": 9,
    "accent_type": 10,
    "accent_con_type": 11,
    "chain_flag": 12,
}

PUNCTUATION_FULL_TO_HALF_TABLE = {
    "、": ",",
    "。": ".",
    "？": "?",
    "！": "!",
}
PUNCTUATION_FULL_TO_HALF_TRANS = str.maketrans(PUNCTUATION_FULL_TO_HALF_TABLE)


def init_seed(seed):
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)


def load_json_corpus(file_path, suffix="json"):
    """Retrieve a list of corpus's path."""
    corpus = []

    if file_path.is_dir():
        file_paths = [path for path in file_path.glob(f"*.{suffix}")]
    else:
        file_paths = [file_path]

    for path in file_paths:
        with open(path, "r") as file:
            corpus += json.load(file)

    return corpus


def split_corpus(
    corpus,
    valid_data_ratio=0.05,
    test_data_ratio=0.05,
    shuffle=True,
    random_state=1,
    absolute_test_size=-1,
):
    """Split corpus into train, valid, test."""
    if absolute_test_size > 0:
        assert absolute_test_size < len(corpus)

        train, valid_test = train_test_split(
            corpus,
            test_size=absolute_test_size,
            shuffle=shuffle,
            random_state=random_state,
        )
        # split valid/test to half size
        valid, test = train_test_split(
            valid_test,
            test_size=0.5,
            shuffle=shuffle,
            random_state=random_state,
        )
    else:
        train, valid_test = train_test_split(
            corpus,
            test_size=(valid_data_ratio + test_data_ratio),
            shuffle=shuffle,
            random_state=random_state,
        )
        valid, test = train_test_split(
            valid_test,
            test_size=(test_data_ratio / (valid_data_ratio + test_data_ratio)),
            shuffle=shuffle,
            random_state=random_state,
        )

    return {"train": train, "val": valid, "test": test}


def sequence_mask(lengths, max_len=None):
    """Compute sequence mask."""
    batch_size = lengths.size(0)

    if max_len is None:
        max_len = lengths.max().item()

    ranges = torch.arange(0, max_len, device=lengths.device).long()
    ranges = ranges.unsqueeze(0).expand(batch_size, max_len)
    ranges = torch.autograd.Variable(ranges)

    lens_exp = lengths.unsqueeze(1).expand_as(ranges)
    mask = ranges < lens_exp

    return mask


def pack_inputs(inputs, embedding_keys, device):
    """Covnert batch to tensor for input"""
    embeddings = {key: inputs[key].to(device) for key in embedding_keys}

    inputs = {
        "embedding_features": embeddings,
        "lengths": inputs["mora_length"].cpu(),
        "mask": sequence_mask(inputs["mora_length"]).to(device),
        "prev_decoder_outputs": {},
    }
    return inputs


def pack_outputs(outputs, device):
    """Covnert batch to tensor for output"""
    outputs = {
        task: {
            "label": outputs[task]["label"].to(device),
            "length": outputs[task]["length"].cpu(),
            "mask": sequence_mask(outputs[task]["length"]).to(device),
        }
        for task in outputs.keys()
    }
    return outputs


def pad_incomplete_accent_logits(logits, target_mask):
    """Add pad token for logit with wrong ap length"""
    if logits.size(1) > target_mask.size(1):
        logits = logits[:, : target_mask.size(1), :]
    elif logits.size(1) < target_mask.size(1):
        _logits = torch.zeros(logits.size(0), target_mask.size(1), logits.size(2))
        _logits[:, : logits.size(1), :] = logits
        logits = _logits.to(logits.device)

    return logits


def get_ap_length(accent_phrases, masks, accent_phrase_bondary_label=2):
    """Count number of accent phrase boundary in sequence.
    Args:
        accent_phrases (tensor): The sequence represents accent phrase.
        mask (tensor): The mask for accent phrase boundary sequence.
        accent_phrase_bondary_label (int, optional):
            The label to represent accnet phrase boundary.
    Returns:
        list: A list of lengths for accent phrase boundary.
    """
    lengths = []

    for accent_phrase, mask in zip(accent_phrases, masks):
        accent_phrase = accent_phrase[mask]
        num_boundary = len(accent_phrase[accent_phrase == accent_phrase_bondary_label])
        lengths.append(num_boundary + 1)

    return lengths


def expand_word_label_to_mora(labels, moras, boundaries, target):
    """Convert word-basd label sequence to mora-based label sequence."""
    mora_labels = []

    for label, mora, boundary in zip(labels, moras, boundaries):
        # Remove unavailable label; because jpbp's positive label means that
        # the break comes after the tagged token,
        # i.e., the label located at last token is meanless
        if target == "intonation_phrase_boundary":
            label[-1] = 0

        label = np.array(label)

        boundary_indexs = np.where(boundary > 0)[0]

        assert len(boundary_indexs) + 1 == len(
            label
        ), f"Not matched length: {len(boundary_indexs) + 1} != {len(label)}"

        if target == "intonation_phrase_boundary":
            label_indexs = np.where(label > 0)[0]
        else:
            raise NotImplementedError(f"Not supported error: {target}")

        # annotate break label
        _label = np.array([LABEL_TABLE[target]["O"]] * len(mora))
        _label[boundary_indexs[label_indexs]] = LABEL_TABLE[target]["B"]

        mora_labels.append(list(_label))

    return mora_labels


def _convert_ap_based_accent_to_mora_based_accent(
    ap_accents, phrases, mode="high_low", mora=None, accent_phrase_boundary_label=2
):
    """Convert accent phrase-based accent status sequence to mora-based."""
    assert mode in ["binary", "high_low"], (
        f"Not supported representation mode {mode}:",
        "Representation mode must be selected in binary and high_low",
    )

    ap_accents = ap_accents.cpu()
    phrases = phrases.cpu()

    boundaries = np.where(phrases == accent_phrase_boundary_label)[0]
    phrases = np.split(phrases, boundaries)

    if mora is not None:
        ap_moras = np.split(np.array(mora), boundaries)

    # Pad or slice if there is miss-macth between predicted phrase and real-label
    if len(phrases) < len(ap_accents):
        ap_accents = ap_accents[: len(phrases)]
    elif len(phrases) > len(ap_accents):
        ap_accents = np.append(ap_accents, [0] * (len(phrases) - len(ap_accents)))

    assert len(phrases) == len(
        ap_accents
    ), f"Not matched seq lengths {len(phrases)} != {len(ap_accents)}"

    mora_accents = []

    for index, (accent, phrase) in enumerate(zip(ap_accents, phrases)):
        if mode == "binary":
            if mora is not None:
                moras = ap_moras[index]
            else:
                moras = ["*"] * len(phrase)

            # convert to 0-based label to index
            accent_label = accent.item() - 2
            mora_accent = [0] * len(phrase)

            if (
                len(mora_accent) > accent_label >= 0
                and moras[accent_label] not in IGNORE_TARGET_MORA_FOR_ACCENT
            ):
                mora_accent[accent_label] = 1
        elif mode == "high_low":
            if mora is not None:
                moras = ap_moras[index]
            else:
                moras = ["*"] * len(phrase)

            # convert to 0-based label to index
            accent_label = accent.item() - 1

            # ignore accent located at out of mora seq
            if accent_label > len(moras):
                accent_label = 0
            # ignore accent located at long-vowel
            elif (
                accent_label > 0
                and moras[accent_label - 1] in IGNORE_TARGET_MORA_FOR_ACCENT
            ):
                accent_label = 0

            _, mora_accent = pron2mora(moras, accent_label, mode)
        else:
            raise NotImplementedError(
                (
                    f"Not supported representation mode {mode}:"
                    " Representation mode must be selected in binary and high_low"
                )
            )

        mora_accents += mora_accent

    return np.array(mora_accents)


def convert_ap_based_accent_to_mora_based_accent(
    accent_statuses,
    accent_phrase_boundaries,
    ap_seq_masks,
    mora_seq_masks,
    accent_represent_mode="binary",
):
    """Convert accent phrase-based accent status sequence to mora-based."""
    mora_accent_statuses = np.array([], dtype=np.int64)

    for accent_status, accent_phrase_boundary, ap_seq_mask, mora_seq_mask in zip(
        accent_statuses,
        accent_phrase_boundaries,
        ap_seq_masks,
        mora_seq_masks,
    ):
        # accent_status: (T_ap) -> (T_ap - mask_ap)
        # ap_seq_mask: (T_ap)
        accent_status = accent_status[ap_seq_mask]

        # accent_phrase_boundary: (T_mora) -> (T_mora, mask_mora)
        # mora_seq_mask: (T_mora)
        accent_phrase_boundary = accent_phrase_boundary[mora_seq_mask]

        # convert to mora-based sequence
        mora_accent_status = _convert_ap_based_accent_to_mora_based_accent(
            accent_status, accent_phrase_boundary, mode=accent_represent_mode
        )

        mora_accent_statuses = np.concatenate(
            [mora_accent_statuses, mora_accent_status]
        )

    return mora_accent_statuses


def convert_mora_jp_to_en(mora):
    phonemes, _, syllable_boundary = mora2phon(mora, [0] * len(mora))
    temp_en_mora = ""
    en_moras = []

    for phoneme, boundary in zip(phonemes, syllable_boundary):
        if phoneme in [".", ",", "?", "!", "N", "cl"]:
            if temp_en_mora:
                en_moras.append(temp_en_mora)
                temp_en_mora = ""
            en_moras.append(phoneme)
            continue

        temp_en_mora += phoneme

        if boundary == 1:
            if has_longvowel(temp_en_mora):
                consonant = temp_en_mora[:-1]
                vowel = temp_en_mora[-1]
                en_moras += [consonant, vowel]
            else:
                en_moras.append(temp_en_mora)
            temp_en_mora = ""

    return en_moras


def plot_attention(attention, xs=None, ys=None):
    fig, ax = plt.subplots()
    attention = attention.cpu().data.numpy().T

    # draw attention
    im = ax.imshow(attention, aspect="auto", origin="lower", interpolation="none")
    fig.colorbar(im, ax=ax)

    # set labels
    if xs:
        ax.set_xticks(list(range(len(xs))))
        ax.set_xticklabels(xs)
        ax.xaxis.set_major_locator(ticker.MultipleLocator(1))

    if ys:
        ax.set_yticks(list(range(len(ys))))
        ax.set_yticklabels(ys)
        ax.yaxis.set_major_locator(ticker.MultipleLocator(1))

    return fig


def plot_batch_attention(
    inputs,
    logits,
    ap_lengths,
    attentions,
    feature_set,
    plot_targets=None,
    tensorboard_writer=None,
    phase=None,
    epoch=None,
    script_ids=None,
    accent_phrase_boundary_label=2,
):
    """Plot attentions in a batch"""
    # parse features
    moras = inputs["embedding_features"]["mora"]
    input_lengths = inputs["lengths"]
    aps = inputs["prev_decoder_outputs"]["accent_phrase_boundary"].cpu().data.numpy()

    if plot_targets:
        indexs = plot_targets
    else:
        indexs = range(len(attentions))

    for index in indexs:
        # skip plotting when got invalid index
        if index >= len(attentions):
            continue

        attention = attentions[index][: ap_lengths[index], : input_lengths[index]]
        accents = torch.argmax(logits[index][: ap_lengths[index]], dim=1).tolist()
        ap = aps[index][: input_lengths[index]].tolist()
        mora = feature_set.convert_id_to_feature(
            "mora", moras[index][: input_lengths[index]].tolist()
        )
        mora = convert_mora_jp_to_en(mora)

        xs = [str(a - 1) for a in accents]
        ys = [
            f"{m}{'/' if a == accent_phrase_boundary_label else '_'}"
            for m, a in zip(mora, ap)
        ]

        fig = plot_attention(attention, xs=xs, ys=ys)

        if tensorboard_writer:
            tensorboard_writer.add_figure(
                f"{phase}/attention/{script_ids[index]}", fig, epoch
            )


def convert_readable_labels(predicts, targets, masks, script_ids):
    """Convert logits to readable label"""
    logs = {}

    for predict, target, mask, script_id in zip(predicts, targets, masks, script_ids):
        predict = [str(int(y)) for y in predict[mask].cpu().tolist()]
        target = [str(int(y)) for y in target[mask].cpu().tolist()]

        logs[script_id] = {
            "predict": ",".join(predict),
            "target": ",".join(target),
            "status": JSON_LOG_STATUS_TO_TAG[predict == target],
        }

    return logs


def group_by_script_id(logs):
    _logs = {}

    for task, scripts in logs.items():
        for script_id, script_info in scripts.items():
            if script_id not in _logs.keys():
                _logs[script_id] = {task: script_info}
            else:
                _logs[script_id][task] = script_info

    return _logs


def _make_task_group_variation(tasks, min_num=2):
    groups = []
    target_size = len(tasks)
    while target_size - min_num >= 0:
        for x in range(0, (len(tasks) - target_size) + 1):
            group = tasks[x : x + target_size]
            groups.append(group)
        target_size -= 1
    return groups


def _calculate_multiple_task_scores(tasks, logs):
    task_groups = _make_task_group_variation(tasks)
    multiple_task_scores = {}

    for task_group in task_groups:
        multiple_task_key = "+".join(task_group)
        statuses = np.array(
            [
                sum(
                    [
                        JSON_LOG_TAG_TO_STATUS[script_info[task]["status"]]
                        for task in task_group
                    ]
                )
                for script_info in logs.values()
            ]
        )
        num_all_task_success = len(np.where(statuses >= len(task_group))[0])
        multiple_task_scores[multiple_task_key] = {
            "sentence_level_accuracy": num_all_task_success / len(logs)
        }

    return multiple_task_scores


def log_scores(
    phase,
    epoch,
    tasks,
    metrics,
    logs=None,
    loss=None,
    tensorboard_writer=None,
):
    """Log scores"""

    scores = metrics.compute()

    # merge loss into score metrics
    if loss:
        for task in tasks:
            scores[task]["loss"] = loss[task].item()

    if logs:
        multiple_task_scores = _calculate_multiple_task_scores(tasks, logs)
        tasks += list(multiple_task_scores.keys())
        scores.update(multiple_task_scores)

    for task in tasks:
        # logging to shell
        score_log = f"{phase} / {task} |" + " |".join(
            [
                f" {score_name} : {score:.4f}"
                if score_name == "loss"
                else f" {score_name} : {score:.4%}"
                for score_name, score in scores[task].items()
            ]
        )
        logger.info(score_log)

        # logging to tensorboard
        if tensorboard_writer:
            for score_name, score in scores[task].items():
                tensorboard_writer.add_scalar(
                    f"{phase}/{task}/{score_name}", score, epoch
                )


def convert_open_jtalk_node_to_feature(nodes):
    features = []
    raw_feature_keys = RAW_FEATURE_KEYS["open-jtalk"]

    for node in nodes:
        # parse feature
        _node = node.split(",")
        node_feature = {}

        for feature_key in raw_feature_keys:
            index = OPEN_JTALK_FEATURE_INDEX_TABLE[feature_key]

            if feature_key == "pos":
                value = ":".join([_node[i] for i in index])
            elif feature_key == "accent_type":
                value = int(_node[index].split("/")[0])
            elif feature_key == "accent_con_type":
                value = _node[index].replace("/", ",")
            elif feature_key == "chain_flag":
                value = int(_node[index])
            elif feature_key == "pron":
                value = _node[index].replace("’", "").replace("ヲ", "オ")
            else:
                value = _node[index]

            node_feature[feature_key] = value

        if node_feature["surface"] == "・":
            continue
        elif node_feature["surface"] in PUNCTUATION_FULL_TO_HALF_TABLE.keys():
            surface = node_feature["surface"].translate(PUNCTUATION_FULL_TO_HALF_TRANS)
            pron = None
            node_feature["surface"] = surface
            node_feature["pron"] = pron

        features.append(node_feature)

    return features


def convert_open_jtalk_format_label(
    labels,
    morph_boundaries,
    accent_nucleus_label=1,
    accent_phrase_boundary_label=1,
    morph_boundary_label=1,
):
    assert "accent_status" in labels.keys(), "`accent_status` is missing in labels"
    assert (
        "accent_phrase_boundary" in labels.keys()
    ), "`accent_phrase_boundary` is missing in labels"

    # squeeze results
    mora_accent_status = labels["accent_status"][0]
    mora_accent_phrase_boundary = labels["accent_phrase_boundary"][0]
    morph_boundary = morph_boundaries[0]

    assert len(mora_accent_status) == len(mora_accent_phrase_boundary), (
        "Not match sequence lenght between"
        "`accent_status`, `morph_boundary`, and `accent_phrase_boundary`"
    )

    mora_accent_phrase_boundary = np.array(mora_accent_phrase_boundary)

    # convert mora-based accent phrase boundary label to morph-based label
    morph_boundary_indexes = np.where(morph_boundary == morph_boundary_label)[0]
    morph_accent_phrase_boundary = np.split(
        mora_accent_phrase_boundary, morph_boundary_indexes
    )
    # `chain_flag` in OpenJTalk represents the status whether the morph will be connected
    morph_accent_phrase_boundary = [
        0 if boundary[0] == accent_phrase_boundary_label else 1
        for boundary in morph_accent_phrase_boundary
    ]
    # first `chain_flag` must be -1
    morph_accent_phrase_boundary[0] = -1
    num_boundary = morph_accent_phrase_boundary.count(0) + 1

    # convert mora-based accent status label to ap-based label
    mora_accent_phrase_boundary_indexes = np.where(
        mora_accent_phrase_boundary == accent_phrase_boundary_label
    )[0]
    phrase_accent_statuses = np.split(
        mora_accent_status, mora_accent_phrase_boundary_indexes
    )
    phrase_accent_status_labels = []

    for phrase_accent_status in phrase_accent_statuses:
        accent_nucleus_indexes = np.where(phrase_accent_status == accent_nucleus_label)[
            0
        ]
        if len(accent_nucleus_indexes) == 0:
            accent_nucleus_index = 0
        else:
            accent_nucleus_index = accent_nucleus_indexes[0] + 1
        phrase_accent_status_labels.append(accent_nucleus_index)

    if len(phrase_accent_status_labels) > num_boundary:
        warnings.warn(
            (
                "Lenght of AP-based accent status will be adjusted "
                "by morph-based accent phrase boundary: "
                f"{len(phrase_accent_status_labels)} > {num_boundary}"
            )
        )
        phrase_accent_status_labels = phrase_accent_status_labels[:num_boundary]

    # convert mora-based accent status to morph-based label
    # the accent label for OpenJTalk pushed in first morph
    morph_accent_status = [
        phrase_accent_status_labels.pop(0) if morph_accent_phrase_flag < 1 else 0
        for morph_accent_phrase_flag in morph_accent_phrase_boundary
    ]

    return {
        "accent_status": morph_accent_status,
        "accent_phrase_boundary": morph_accent_phrase_boundary,
    }



from logging import getLogger

from hydra.utils import instantiate

logger = getLogger(__name__)


def init_model(tasks, config, feature_set, device, is_train=False):
    if is_train:
        criterions = {}
        optimizer = None
        scheduler = None

    # setting shared layers
    # Embedding
    embedding_kwargs = {"feature_set": feature_set}
    embedding = instantiate(config.model.embedding, **embedding_kwargs)

    # Encoder
    encooder_input_size = sum(
        [config.model.embedding.embeding_sizes[key] for key in config.data.input_keys]
    )

    encoder_output_size = config.model.encoder.param.hidden_size * 2
    encoder_kwargs = {"input_size": encooder_input_size}

    encoders = {}
    for task in tasks:
        if (
            config.model.encoder.shared_with[task]
            and config.model.encoder.shared_with[task] in encoders.keys()
        ):
            logger.info(
                f"{task} has shared encoder with {config.model.encoder.shared_with[task]}"
            )
            encoders[task] = encoders[config.model.encoder.shared_with[task]]
        else:
            encoder = instantiate(config.model.encoder.param, **encoder_kwargs)
            encoders[task] = encoder

    # Decoder
    decoder_input_sizes = [encoder_output_size] * len(tasks)

    decoder_kwargs = {
        task: {
            "input_size": decoder_input_size,
            "output_size": config.data.output_sizes[task],
        }
        for task, decoder_input_size in zip(tasks, decoder_input_sizes)
    }

    # init decoders with criterion
    decoders = {}
    for task in tasks:
        # init embedding for
        if config.model.decoder[task]["prev_task_embedding_label_list"]:
            assert config.model.decoder[task]["prev_task_embedding_label_list"] == list(
                config.model.decoder[task]["prev_task_embedding_size"].keys()
            ), "Not matched embedding setting for previous tasks: {} != {}".format(
                config.model.decoder[task]["prev_task_embedding_label_list"],
                config.model.decoder[task]["prev_task_embedding_size"].keys(),
            )

            decoder_kwargs[task]["prev_task_embedding_label_size"] = {
                task_label: decoder_kwargs[task_label]["output_size"]
                for task_label in config.model.decoder[task][
                    "prev_task_embedding_label_list"
                ]
                if task_label in tasks
            }

        decoders[task] = instantiate(config.model.decoder[task], **decoder_kwargs[task])

        if is_train:
            if config.criterions[task]._target_ == "marine.criterions.LogLikelhood":
                criterion_kwargs = {"log_likehood_func": decoders[task].crf}
            else:
                criterion_kwargs = {}

            criterion = instantiate(config.criterions[task], **criterion_kwargs).to(
                device
            )
            criterions[task] = criterion

    # Init base model
    model_kwarwgs = {"embedding": embedding, "encoders": encoders, "decoders": decoders}

    model = instantiate(config.model.base, **model_kwarwgs).to(device)

    logger.debug(f"model has initialized\n{model}")

    if is_train:
        optimizer_kwargs = {"params": model.parameters()}
        optimizer = instantiate(config.optim.optimizer, **optimizer_kwargs)

        scheduler_kwargs = {"optimizer": optimizer}
        scheduler = instantiate(config.optim.scheduler, **scheduler_kwargs)

        return model, criterions, optimizer, scheduler
    else:
        return model



from .accent import (
    represent_accent_binary,
    represent_accent_high_low,
    represent_longvowel_accent_binary,
    represent_longvowel_accent_high_low,
    set_accent_status,
)
from .boundary import represent_syllable_boundary
from .util import (
    CONNECTABLE_MORA,
    HALF_PUNCTUATION,
    LONGVOWEL_CHARACTER,
    NON_MORA_LIST,
    SUPPORTED_MORA,
    get_phoneme,
)

ACCENT_REPRESENT_FUNC_TABLE = {
    "binary": {
        "represent_accent": represent_accent_binary,
        "represent_longvowel_accent": represent_longvowel_accent_binary,
    },
    "high_low": {
        "represent_accent": represent_accent_high_low,
        "represent_longvowel_accent": represent_longvowel_accent_high_low,
    },
}


def pron2mora(pron, accent=None, represent_mode="binary"):
    moras = []
    i = 0

    while i < len(pron):
        current_pron = pron[i]

        if current_pron in NON_MORA_LIST and len(moras) > 0:
            merged_mora = f"{moras[-1]}{current_pron}"

            if merged_mora in SUPPORTED_MORA:
                moras[-1] = merged_mora
            else:
                moras.append(current_pron)
        else:
            moras.append(current_pron)

        i += 1

    if accent is not None:
        if not isinstance(accent, int):
            raise TypeError(f"Accent is must be int not {type(accent)}")

        if represent_mode not in ACCENT_REPRESENT_FUNC_TABLE.keys():
            raise NotImplementedError(f"Not Implemented mode : {represent_mode}")

        # init rule
        accent_rule = ACCENT_REPRESENT_FUNC_TABLE[represent_mode]
        represent_accent = accent_rule["represent_accent"]
        represent_longvowel_accent = accent_rule["represent_longvowel_accent"]

        # init satus
        high, end_low = set_accent_status(accent)
        represented_accents = []

        for index, mora in enumerate(moras):
            # if currnet mora is long-vowel syombol, update last mora
            if _is_longvowel(mora):
                represented_accent = represent_longvowel_accent(index, high, end_low)
                represented_accents.append(represented_accent)
            else:
                represented_accent = represent_accent(index, high, end_low)
                represented_accents.append(represented_accent)

        assert len(moras) == len(
            represented_accents
        ), f"Wrong repersentation : {moras} != {represented_accents}"

        return moras, represented_accents

    return moras


def _is_longvowel(mora):
    return mora == LONGVOWEL_CHARACTER


# Consider whether the mora is long-vowel
# and previous mora was not unaccneted mora for escapte excepted case
# e.g., ンー = NN11, ッー = cl11
def _is_prev_mora_not_unaccented_mora(index, moras):
    return index > 0 and moras[index - 1] not in CONNECTABLE_MORA


def mora2phon(
    moras,
    accents=None,
    ignore_longvowel_accent=False,
    use_punctuation=True,
    punctuation_accent_label=7,
):
    phonemes = []

    if accents is None:
        for mora in moras:
            if not use_punctuation and mora in HALF_PUNCTUATION:
                continue

            phoneme = get_phoneme(mora, phonemes)
            phonemes += phoneme

        return phonemes

    else:
        if len(accents) != len(moras):
            raise ValueError(
                f"Accent is must be same to length of mora (got : {len(accents) != len(moras)})"
            )

        represented_accents = []
        represented_boundaries = []

        for index, (mora, accent) in enumerate(zip(moras, accents)):
            if not use_punctuation and mora in HALF_PUNCTUATION:
                continue

            phoneme = get_phoneme(mora, phonemes)
            len_phoneme = len(phoneme)
            represented_boundary = represent_syllable_boundary(
                index, moras, len_phoneme
            )
            phonemes += phoneme

            # if currnet mora is long-vowel syombol, update last mora
            if _is_longvowel(mora):
                # if currnet mora is long-vowel symbol, only update previous accent
                if not ignore_longvowel_accent and _is_prev_mora_not_unaccented_mora(
                    index, moras
                ):
                    represented_accents[-1] = accent
                represented_boundaries[-1] = represented_boundary[-1]
            elif mora in HALF_PUNCTUATION:
                represented_accent = [punctuation_accent_label]
                represented_accents += represented_accent
                represented_boundaries += represented_boundary
            else:
                represented_accent = [0] * (len_phoneme - 1) + [accent]
                represented_accents += represented_accent
                represented_boundaries += represented_boundary

        return phonemes, represented_accents, represented_boundaries


def pron2phon(pron, accent=None, represent_mode="binary"):
    if represent_mode not in ACCENT_REPRESENT_FUNC_TABLE.keys():
        raise NotImplementedError(f"Not Implemented mode : {represent_mode}")

    if accent is None:
        moras = pron2mora(pron, accent, represent_mode)
        accents = None
    else:
        moras, accents = pron2mora(pron, accent, represent_mode)

    return mora2phon(moras, accents)



def set_accent_status(accent):
    high = -1
    end_low = -1

    # 1-type : H-L-L ...
    if accent == 1:
        high = 0
        end_low = 1

    # 0-type : L-H-H ...
    elif accent <= 0:
        high = 1

    # N-type : L-H ... H_n-L_n+1-L_n+2 ..
    else:
        high = 1
        end_low = accent

    return high, end_low


def represent_accent_high_low(index, high, end_low):
    """
    Represent the accent by a current status of the mora
    - types:
        - 0: Low
        - 1: High
    """

    # Init accent into flat-accent(=0)
    accent = 0

    # set current accent to high
    if index >= high and (end_low < 0 or index < end_low):
        accent = 1

    return accent


def represent_longvowel_accent_high_low(index, high, end_low):
    # Init accent into flat accent
    accent = 0

    # make previous accent to high
    if index >= high and (end_low < 0 or index <= end_low):
        accent = 1

    return accent


def represent_accent_binary(index, high, end_low):
    """
    Represent the accent by a current status of the mora
    - types:
        - 0: Not accent nucleus
        - 1: Accent nucleus
    """

    # Init accent into flat-accent(=0), and make previous accent to high if match it
    accent = 1 if index == end_low - 1 else 0

    return accent


def represent_longvowel_accent_binary(index, high, end_low):
    # Init accent into a flat accent, and make previous accent to high if match it
    return 1 if index == end_low else 0



import torch
from marine.modules.attention import BahdanauAttention, ZoneOutCell
from marine.utils.util import get_ap_length
from torch import nn


class AttentionBasedLSTMDecoder(nn.Module):
    def __init__(
        self,
        input_size=512,
        output_size=20,
        hidden_size=512,
        num_layers=2,
        attention_hidden_size=128,
        decoder_embedding_size=256,
        zoneout=0.1,
        prev_task_dropout=0.5,
        decoder_prev_out_dropout=0.5,
        prev_task_embedding_label_list=None,
        prev_task_embedding_label_size=None,
        prev_task_embedding_size=None,
        padding_idx=0,
    ):
        super().__init__()
        # NOTE: output_size must includes size for [PAD]
        self.output_size = output_size

        if prev_task_embedding_label_size:
            embeddings = {}
            dropouts = {}
            for key in prev_task_embedding_label_list:
                embeddings[key] = nn.Embedding(
                    prev_task_embedding_label_size[key],
                    prev_task_embedding_size[key],
                    padding_idx=padding_idx,
                )
                input_size += prev_task_embedding_size[key]

                if prev_task_dropout:
                    dropouts[key] = nn.Dropout(prev_task_dropout)

            self.prev_task_embedding = nn.ModuleDict(embeddings)

            if len(dropouts) > 0:
                self.prev_task_dropout = nn.ModuleDict(dropouts)
            else:
                self.prev_task_dropout = None

        else:
            self.prev_task_embedding = None
            self.prev_task_dropout = None

        self.attention = BahdanauAttention(
            input_size,
            hidden_size,
            attention_hidden_size,
        )

        # in_dim: output-size + [PAD + SOS]
        self.decoder_embedding = nn.Embedding(
            self.output_size + 1, decoder_embedding_size, padding_idx=padding_idx
        )

        self.decoder_prev_out_dropout = nn.Dropout(decoder_prev_out_dropout)

        # Setup autogressive LSTM layer
        self.lstm = nn.ModuleList()
        for layer in range(num_layers):
            lstm = nn.LSTMCell(
                input_size + decoder_embedding_size if layer == 0 else hidden_size,
                hidden_size,
            )
            self.lstm += [ZoneOutCell(lstm, zoneout)]

        # Setup feature projection layer
        project_size = input_size + hidden_size
        # out_dim: output-size + [PAD]
        self.projection = nn.Linear(project_size, self.output_size, bias=False)

    def _zero_state(self, hs):
        init_hs = hs.new_zeros(hs.size(0), self.lstm[0].hidden_size)
        return init_hs

    def forward(self, encoder_outputs, mask, prev_task_outputs, decoder_targets=None):
        is_inference = decoder_targets is None
        ap_lengths = get_ap_length(prev_task_outputs["accent_phrase_boundary"], mask)

        if is_inference:
            max_decoder_time_steps = max(ap_lengths)
        else:
            max_decoder_time_steps = decoder_targets.shape[1]

        if self.prev_task_embedding is not None:
            prev_task_output_embs = []
            for key in self.prev_task_embedding.keys():
                prev_task_output = prev_task_outputs[key]
                prev_task_output_emb = self.prev_task_embedding[key](prev_task_output)

                if self.prev_task_dropout:
                    prev_task_output_emb = self.prev_task_dropout[key](
                        prev_task_output_emb
                    )

                prev_task_output_embs.append(prev_task_output_emb)

            encoder_outputs = torch.cat(
                [encoder_outputs] + prev_task_output_embs, dim=2
            )

        h_list, c_list = [], []
        for _ in range(len(self.lstm)):
            h_list.append(self._zero_state(encoder_outputs))
            c_list.append(self._zero_state(encoder_outputs))

        go_frame = encoder_outputs.new_zeros(encoder_outputs.size(0), dtype=torch.long)
        go_frame[:] = self.output_size  # SOS
        prev_out = go_frame

        self.attention.reset()

        outs, att_ws = [], []
        t = 0

        while True:
            att_c, att_w = self.attention(encoder_outputs, h_list[0], mask)

            decoder_emb = self.decoder_embedding(prev_out)
            decoder_emb = self.decoder_prev_out_dropout(decoder_emb)

            # LSTM
            xs = torch.cat([att_c, decoder_emb], dim=1)
            h_list[0], c_list[0] = self.lstm[0](xs, (h_list[0], c_list[0]))
            for i in range(1, len(self.lstm)):
                h_list[i], c_list[i] = self.lstm[i](
                    h_list[i - 1], (h_list[i], c_list[i])
                )
            hcs = torch.cat([h_list[-1], att_c], dim=1)

            outs.append(
                # (B, out_dim) -> (B, 1, out_dim)
                self.projection(hcs).view(
                    encoder_outputs.size(0),
                    -1,
                    self.output_size,
                )
            )
            att_ws.append(att_w)

            if is_inference:
                # List[(B, 1, out_dim)] -> (B, out_dim) -> (B, 1)
                prev_out = torch.argmax(outs[-1][:, -1, :], dim=1)
            else:
                # Teacher forcing
                # (B, Lmax) -> (B, 1)
                prev_out = decoder_targets[:, t]

            t += 1
            if t >= max_decoder_time_steps:
                break

        outs = torch.cat(outs, dim=1)  # (B, Lmax, out_dim)
        att_ws = torch.stack(att_ws, dim=1)  # (B, Lmax, Tmax)

        return outs, att_ws, ap_lengths



import torch
from torch import nn
from torch.nn import functional as F


class ZoneOutCell(nn.Module):
    def __init__(self, cell, zoneout=0.1):
        super().__init__()
        self.cell = cell
        self.hidden_size = cell.hidden_size
        self.zoneout = zoneout

    def forward(self, inputs, hidden):
        next_hidden = self.cell(inputs, hidden)
        next_hidden = self._zoneout(hidden, next_hidden, self.zoneout)
        return next_hidden

    def _zoneout(self, h, next_h, prob):
        h_0, c_0 = h
        h_1, c_1 = next_h
        h_1 = self._apply_zoneout(h_0, h_1, prob)
        c_1 = self._apply_zoneout(c_0, c_1, prob)
        return h_1, c_1

    def _apply_zoneout(self, h, next_h, prob):
        if self.training:
            mask = h.new(*h.size()).bernoulli_(prob)
            return mask * h + (1 - mask) * next_h
        else:
            return prob * h + (1 - prob) * next_h


class BahdanauAttention(nn.Module):
    """Bahdanau-style attention
    This is an attention mechanism originally used in Tacotron.
    Args:
        encoder_dim (int): dimension of encoder outputs
        decoder_dim (int): dimension of decoder outputs
        hidden_dim (int): dimension of hidden state
    """

    def __init__(self, encoder_dim=512, decoder_dim=1024, hidden_dim=128):
        super().__init__()
        self.mlp_enc = nn.Linear(encoder_dim, hidden_dim)
        self.mlp_dec = nn.Linear(decoder_dim, hidden_dim, bias=False)
        self.w = nn.Linear(hidden_dim, 1)

        self.processed_memory = None

    def reset(self):
        """Reset the internal buffer"""
        self.processed_memory = None

    def forward(
        self,
        encoder_outs,
        decoder_state,
        mask=None,
    ):
        """Forward step
        Args:
            encoder_outs (torch.FloatTensor): encoder outputs
            src_lens (list): length of each input batch
            decoder_state (torch.FloatTensor): decoder hidden state
            mask (torch.FloatTensor): mask for padding
        """

        if self.processed_memory is None:
            self.processed_memory = self.mlp_enc(encoder_outs)

        decoder_state = self.mlp_dec(decoder_state).unsqueeze(1)

        erg = self.w(torch.tanh(self.processed_memory + decoder_state)).squeeze(-1)

        if mask is not None:
            # invert mask
            mask = ~mask
            erg.masked_fill_(mask, -float("inf"))

        attention_weights = F.softmax(erg, dim=1)

        attention_context = torch.sum(
            encoder_outs * attention_weights.unsqueeze(-1), dim=1
        )

        return attention_context, attention_weights



# flake8: noqa: F401
from .att_lstm_decoder import *
from .base_model import *
from .bilstm_encoder import *
from .crf_decoder import *
from .embedding import *
from .linear_decoder import *
from .util import *



import pytest
from marine.predict import Predictor


@pytest.fixture
def predictor() -> Predictor:
    """load inference model using default config"""
    return Predictor()


def test_predict(predictor):
    """just to confirm predict() is working without errors."""
    nodes = [
        {
            "surface": "水",
            "pron": "ミズ",
            "pos": "名詞:一般:*:*",
            "c_type": "*",
            "c_form": "*",
            "accent_type": 0,
            "accent_con_type": "C3",
            "chain_flag": -1,
        },
        {
            "surface": "を",
            "pron": "オ",
            "pos": "助詞:格助詞:一般:*",
            "c_type": "*",
            "c_form": "*",
            "accent_type": 0,
            "accent_con_type": "動詞%F5,名詞%F1",
            "chain_flag": 1,
        },
        {
            "surface": "マレーシア",
            "pron": "マレーシア",
            "pos": "名詞:固有名詞:地域:国",
            "c_type": "*",
            "c_form": "*",
            "accent_type": 2,
            "accent_con_type": "C1",
            "chain_flag": 0,
        },
        {
            "surface": "から",
            "pron": "カラ",
            "pos": "助詞:格助詞:一般:*",
            "c_type": "*",
            "c_form": "*",
            "accent_type": 2,
            "accent_con_type": "名詞%F1",
            "chain_flag": 1,
        },
        {
            "surface": "買わ",
            "pron": "カワ",
            "pos": "動詞:自立:*:*",
            "c_type": "五段・ワ行促音便",
            "c_form": "未然形",
            "accent_type": 0,
            "accent_con_type": "*",
            "chain_flag": 0,
        },
        {
            "surface": "なく",
            "pron": "ナク",
            "pos": "助動詞:*:*:*",
            "c_type": "特殊・ナイ",
            "c_form": "連用テ接続",
            "accent_type": 1,
            "accent_con_type": "動詞%F3@0",
            "chain_flag": 1,
        },
        {
            "surface": "て",
            "pron": "テ",
            "pos": "助詞:接続助詞:*:*",
            "c_type": "*",
            "c_form": "*",
            "accent_type": 0,
            "accent_con_type": "動詞%F1,形容詞%F1,名詞%F5",
            "chain_flag": 1,
        },
        {
            "surface": "は",
            "pron": "ワ",
            "pos": "助詞:係助詞:*:*",
            "c_type": "*",
            "c_form": "*",
            "accent_type": 0,
            "accent_con_type": "名詞%F1,動詞%F2@0,形容詞%F2@0",
            "chain_flag": 1,
        },
        {
            "surface": "なら",
            "pron": "ナラ",
            "pos": "動詞:非自立:*:*",
            "c_type": "五段・ラ行",
            "c_form": "未然形",
            "accent_type": 2,
            "accent_con_type": "*",
            "chain_flag": 0,
        },
        {
            "surface": "ない",
            "pron": "ナイ",
            "pos": "助動詞:*:*:*",
            "c_type": "特殊・ナイ",
            "c_form": "基本形",
            "accent_type": 1,
            "accent_con_type": "動詞%F3@0,形容詞%F2@1",
            "chain_flag": 1,
        },
        {
            "surface": "の",
            "pron": "ノ",
            "pos": "名詞:非自立:一般:*",
            "c_type": "*",
            "c_form": "*",
            "accent_type": 2,
            "accent_con_type": "動詞%F2@0,形容詞%F2@-1",
            "chain_flag": 0,
        },
        {
            "surface": "です",
            "pron": "デス",
            "pos": "助動詞:*:*:*",
            "c_type": "特殊・デス",
            "c_form": "基本形",
            "accent_type": 1,
            "accent_con_type": "名詞%F2@1,動詞%F1,形容詞%F2@0",
            "chain_flag": 1,
        },
        {
            "surface": ".",
            "pron": None,
            "pos": "記号:句点:*:*",
            "c_type": "*",
            "c_form": "*",
            "accent_type": 0,
            "accent_con_type": "*",
            "chain_flag": 0,
        },
    ]
    feature = predictor.predict([nodes])

    print(feature)



from torch import cat, nn


class SimpleEmbedding(nn.Module):
    def __init__(self, embeding_sizes, dropout, feature_set):
        super().__init__()

        # embeddings
        self.embeddings = nn.ModuleDict(
            {
                key: nn.Embedding(
                    len(feature_set.feature_to_id[key]),
                    embeding_sizes[key],
                    padding_idx=feature_set.feature_to_id[key][feature_set.pad_token],
                )
                for key in feature_set.feature_keys
            }
        )

        if dropout:
            self.dropout = nn.Dropout(dropout)
        else:
            self.dropout = None

    def forward(self, **kwargs):
        # Embedding -> B * T * Embedding-size
        embs = [self.embeddings[key](kwargs[key]) for key in self.embeddings.keys()]
        embs = cat(embs, dim=2)

        if self.dropout:
            embs = self.dropout(embs)

        return embs



import argparse
import json
import sys
from concurrent.futures import ProcessPoolExecutor
from multiprocessing import cpu_count
from pathlib import Path

from marine.logger import getLogger
from marine.utils.util import convert_open_jtalk_node_to_feature, load_json_corpus
from tqdm import tqdm

logger = None


def get_parser():
    parser = argparse.ArgumentParser(
        description="Convert Special format txt format data to json file",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument("in_path", type=Path, help="Input path or directory")
    parser.add_argument("out_dir", type=Path, help="Output directory")
    parser.add_argument("--n_jobs", type=int, default=8, help="Number of jobs")
    parser.add_argument(
        "--verbose",
        "-v",
        type=int,
        default=50,
        help="Logging level",
    )
    return parser


def extract_feature(script_id, text):
    features = {"script_id": script_id, "nodes": []}

    try:
        from pyopenjtalk import run_frontend
    except BaseException:
        raise ImportError(
            'Please install pyopenjtalk by `pip install -e ".[dev,pyopenjtalk]"`'
        )

    # drop full-context label
    nodes, _ = run_frontend(text)
    features["nodes"] = convert_open_jtalk_node_to_feature(nodes)

    return features


def _sort_corpus_by_script_id(corpus):
    return list(sorted(corpus, key=lambda x: x["script_id"]))


def entry(argv=sys.argv):
    global logger

    args = get_parser().parse_args(argv[1:])
    logger = getLogger(args.verbose)
    logger.debug(args)

    # Process
    n_jobs = min(cpu_count(), args.n_jobs)

    if not args.out_dir.exists():
        args.out_dir.mkdir(parents=True)

    # Load corpus
    corpus = load_json_corpus(args.in_path)

    if n_jobs > 1:
        logger.info(f"Processing {len(corpus):,} scripts with {n_jobs} jobs")
        with ProcessPoolExecutor(n_jobs) as executor:
            futures = [
                executor.submit(
                    extract_feature,
                    script["script_id"],
                    script["surface"],
                )
                for script in corpus
            ]
            corpus = [
                future.result()
                for future in tqdm(
                    futures, desc="Convert corpus to feature", leave=False
                )
            ]
    else:
        logger.info(f"Processing {len(corpus):,} scripts in a single thread")
        corpus = [
            extract_feature(script["script_id"], script["surface"]) for script in corpus
        ]

    corpus = _sort_corpus_by_script_id(corpus)

    output_path = args.out_dir / f"{args.in_path.parent.name}.json"

    with open(output_path, "w") as file:
        json.dump(corpus, file, ensure_ascii=False, indent=4, separators=(",", ": "))


if __name__ == "__main__":
    sys.exit(entry())



from torch import cat, nn


class LinearDecoder(nn.Module):
    def __init__(
        self,
        input_size,
        output_size,
        prev_task_embedding_label_list=None,
        prev_task_embedding_label_size=None,
        prev_task_embedding_size=None,
        prev_task_dropout=None,
        padding_idx=0,
    ):
        super().__init__()
        if prev_task_embedding_label_size:
            embeddings = {}
            dropouts = {}

            for key in prev_task_embedding_label_list:
                embeddings[key] = nn.Embedding(
                    prev_task_embedding_label_size[key],
                    prev_task_embedding_size[key],
                    padding_idx=padding_idx,
                )
                input_size += prev_task_embedding_size[key]

                if prev_task_dropout:
                    dropouts[key] = nn.Dropout(prev_task_dropout)

            self.prev_task_embedding = nn.ModuleDict(embeddings)

            if len(dropouts) > 0:
                self.prev_task_dropout = nn.ModuleDict(dropouts)
            else:
                self.prev_task_dropout = None

        else:
            self.prev_task_embedding = None
            self.prev_task_dropout = None

        # NOTE: output_size must includes size for [PAD]
        self.linear = nn.Linear(input_size, output_size, bias=True)

    def forward(self, logits, mask, prev_decoder_outputs=None, decoder_targets=None):
        if self.prev_task_embedding is not None:
            prev_decoder_output_embs = []

            for key in self.prev_task_embedding.keys():
                prev_decoder_output = prev_decoder_outputs[key]
                prev_decoder_output_emb = self.prev_task_embedding[key](
                    prev_decoder_output
                )

                if self.prev_task_dropout:
                    prev_decoder_output_emb = self.prev_task_dropout[key](
                        prev_decoder_output_emb
                    )

                prev_decoder_output_embs.append(prev_decoder_output_emb)

            logits = cat([logits] + prev_decoder_output_embs, dim=2)

        # Linear -> B * T * Output-size
        linear_logits = self.linear(logits)

        return linear_logits



"""
Feature tables for embeddings
"""

import re

# Mora list
MORAS = [
    # x
    "ア",
    "イ",
    "ウ",
    "エ",
    "オ",
    "ワ",
    # w
    "ウィ",
    "ウェ",
    "ウォ",
    # y
    "ヤ",
    "ユ",
    "ヨ",
    "イェ",
    # k
    "カ",
    "キ",
    "ク",
    "ケ",
    "コ",
    # kw
    "クァ",
    "クィ",
    "クェ",
    "クォ",
    "クヮ",
    # ky
    "キャ",
    "キュ",
    "キェ",
    "キョ",
    # g
    "ガ",
    "ギ",
    "グ",
    "ゲ",
    "ゴ",
    # gw
    "グァ",
    "グィ",
    "グェ",
    "グォ",
    "グヮ",
    # gy
    "ギャ",
    "ギュ",
    "ギェ",
    "ギョ",
    # s
    "サ",
    "スィ",
    "ス",
    "セ",
    "ソ",
    # sh
    "シ",
    "シェ",
    "シャ",
    "シュ",
    "ショ",
    # z
    "ザ",
    "ズィ",
    "ズ",
    "ゼ",
    "ゾ",
    "ヂェ",
    # j
    "ジ",
    "ジェ",
    "ジャ",
    "ジュ",
    "ジョ",
    "ヂャ",
    "ヂュ",
    "ヂョ",
    # t
    "タ",
    "ティ",
    "トゥ",
    "テ",
    "ト",
    # ty
    "テャ",
    "テュ",
    "テョ",
    # d
    "ダ",
    "ディ",
    "ドゥ",
    "デ",
    "ド",
    # dy
    "デャ",
    "デュ",
    "デョ",
    # ch
    "チ",
    "チェ",
    "チャ",
    "チュ",
    "チョ",
    # ts
    "ツァ",
    "ツィ",
    "ツ",
    "ツェ",
    "ツォ",
    # n
    "ナ",
    "ニ",
    "ヌ",
    "ネ",
    "ノ",
    # ny
    "ニャ",
    "ニュ",
    "ニェ",
    "ニョ",
    # h
    "ハ",
    "ヒ",
    "ヘ",
    "ホ",
    # hy
    "ヒャ",
    "ヒュ",
    "ヒェ",
    "ヒョ",
    # p
    "パ",
    "ピ",
    "プ",
    "ペ",
    "ポ",
    # pw
    "プァ",
    "プィ",
    "プェ",
    "プォ",
    # py
    "ピャ",
    "ピュ",
    "ピェ",
    "ピョ",
    "プャ",
    "プュ",
    "プョ",
    # b
    "バ",
    "ビ",
    "ブ",
    "ベ",
    "ボ",
    "ヴ",
    # v
    "ヴァ",
    "ヴィ",
    "ヴェ",
    "ヴォ",
    "ブァ",
    "ブィ",
    "ブェ",
    "ブォ",
    # by
    "ビャ",
    "ビュ",
    "ビェ",
    "ビョ",
    # vy
    "ヴャ",
    "ヴュ",
    "ヴョ",
    "ブャ",
    "ブュ",
    "ブョ",
    # f
    "ファ",
    "フィ",
    "フ",
    "フェ",
    "フォ",
    # fy
    "フャ",
    "フュ",
    "フョ",
    # m
    "マ",
    "ミ",
    "ム",
    "メ",
    "モ",
    # my
    "ミャ",
    "ミュ",
    "ミェ",
    "ミョ",
    # r
    "ラ",
    "リ",
    "ル",
    "レ",
    "ロ",
    # ry
    "リャ",
    "リュ",
    "リェ",
    "リョ",
    # N
    "ン",
    # cl
    "ッ",
    # Longvowel symbol
    "ー",
]
PUNCTUATIONS = [",", ".", "?", "!"]
MORAS += PUNCTUATIONS

# POS list defined in Unidic-CSJ
POS_UNIDIC_CSJ = [
    "代名詞:*:*:*",
    "副詞:*:*:*",
    "助動詞:*:*:*",
    # PARTICLE
    "助詞:係助詞:*:*",
    "助詞:副助詞:*:*",
    "助詞:接続助詞:*:*",
    "助詞:格助詞:*:*",
    "助詞:準体助詞:*:*",
    "助詞:終助詞:*:*",
    # VERB
    "動詞:一般:*:*",
    "動詞:非自立可能:*:*",
    # NOUN
    "名詞:助動詞語幹:*:*",
    "名詞:数詞:*:*",
    "名詞:普通名詞:サ変可能:*",
    "名詞:普通名詞:サ変形状詞可能:*",
    "名詞:普通名詞:一般:*",
    "名詞:普通名詞:副詞可能:*",
    "名詞:普通名詞:助数詞可能:*",
    "名詞:普通名詞:形状詞可能:*",
    # NOUN: NER
    "名詞:固有名詞:一般:*",
    "名詞:固有名詞:人名:一般",
    "名詞:固有名詞:人名:名",
    "名詞:固有名詞:人名:姓",
    "名詞:固有名詞:地名:一般",
    "名詞:固有名詞:地名:国",
    # ADJECTIVE
    "形容詞:一般:*:*",
    "形容詞:非自立可能:*:*",
    # NA-ADJECTIVE
    "形状詞:タリ:*:*",
    "形状詞:一般:*:*",
    "形状詞:助動詞語幹:*:*",
    # INTERJECTION
    "感動詞:フィラー:*:*",
    "感動詞:一般:*:*",
    # SUFFIX
    "接尾辞:動詞的:*:*",
    "接尾辞:名詞的:サ変可能:*",
    "接尾辞:名詞的:サ変形状詞可能:*",
    "接尾辞:名詞的:一般:*",
    "接尾辞:名詞的:副詞可能:*",
    "接尾辞:名詞的:助数詞:*",
    "接尾辞:名詞的:形状詞可能:*",
    "接尾辞:形容詞的:*:*",
    "接尾辞:形状詞的:*:*",
    # CONJUNCTION
    "接続詞:*:*:*",
    # PREFIX
    "接頭辞:*:*:*",
    # ADNOMINAL
    "連体詞:*:*:*",
    # SYMBOL
    "空白:*:*:*",
    "補助記号:一般:*:*",
    "補助記号:句点:*:*",
    "補助記号:括弧閉:*:*",
    "補助記号:括弧開:*:*",
    "補助記号:絵文字:一般:*",
    "補助記号:読点:*:*",
    "補助記号:ＡＡ:一般:*",
    "補助記号:ＡＡ:顔文字:*",
    "記号:一般:*:*",
    "記号:文字:*:*",
]

# POS list defined in OpenJTalk dictionary (NAIST-jdic)
POS_OPEN_JTALK = [
    "その他:間投:*:*",
    "フィラー:*:*:*",
    "感動詞:*:*:*",
    "記号:アルファベット:*:*",
    "記号:一般:*:*",
    "記号:括弧開:*:*",
    "記号:括弧閉:*:*",
    "記号:句点:*:*",
    "記号:空白:*:*",
    "記号:読点:*:*",
    "形容詞:自立:*:*",
    "形容詞:接尾:*:*",
    "形容詞:非自立:*:*",
    "助詞:格助詞:一般:*",
    "助詞:格助詞:引用:*",
    "助詞:格助詞:連語:*",
    "助詞:係助詞:*:*",
    "助詞:終助詞:*:*",
    "助詞:接続助詞:*:*",
    "助詞:特殊:*:*",
    "助詞:副詞化:*:*",
    "助詞:副助詞:*:*",
    "助詞:副助詞／並立助詞／終助詞:*:*",
    "助詞:並立助詞:*:*",
    "助詞:連体化:*:*",
    "助動詞:*:*:*",
    "接続詞:*:*:*",
    "接頭詞:形容詞接続:*:*",
    "接頭詞:数接続:*:*",
    "接頭詞:動詞接続:*:*",
    "接頭詞:名詞接続:*:*",
    "動詞:自立:*:*",
    "動詞:接尾:*:*",
    "動詞:非自立:*:*",
    "副詞:一般:*:*",
    "副詞:助詞類接続:*:*",
    "名詞:サ変接続:*:*",
    "名詞:ナイ形容詞語幹:*:*",
    "名詞:一般:*:*",
    "名詞:引用文字列:*:*",
    "名詞:形容動詞語幹:*:*",
    "名詞:固有名詞:一般:*",
    "名詞:固有名詞:人名:一般",
    "名詞:固有名詞:人名:姓",
    "名詞:固有名詞:人名:名",
    "名詞:固有名詞:組織:*",
    "名詞:固有名詞:地域:一般",
    "名詞:固有名詞:地域:国",
    "名詞:数:*:*",
    "名詞:接続詞的:*:*",
    "名詞:接尾:サ変接続:*",
    "名詞:接尾:一般:*",
    "名詞:接尾:形容動詞語幹:*",
    "名詞:接尾:助数詞:*",
    "名詞:接尾:助動詞語幹:*",
    "名詞:接尾:人名:*",
    "名詞:接尾:地域:*",
    "名詞:接尾:特殊:*",
    "名詞:接尾:副詞可能:*",
    "名詞:代名詞:一般:*",
    "名詞:代名詞:縮約:*",
    "名詞:動詞非自立的:*:*",
    "名詞:特殊:助動詞語幹:*",
    "名詞:非自立:一般:*",
    "名詞:非自立:形容動詞語幹:*",
    "名詞:非自立:助動詞語幹:*",
    "名詞:非自立:副詞可能:*",
    "名詞:副詞可能:*:*",
    "連体詞:*:*:*",
]

# Accent-con-type defined in Unidic-CSJ
ACCENT_CON_TYPES_UNIDIC_CSJ = [
    "1",
    "C1",
    "C2",
    "C3",
    "C4",
    "C5",
    "F1",
    "F2",
    "F3",
    "F4",
    "F6",
    "P1",
    "P2",
    "P4",
    "P6",
    "P13",
    "P14",
]

# Accent-con-type defined in OpenJTalk dictionary (NAIST-jdic)
ACCENT_CON_TYPES_OPEN_JTALK = [
    "1",
    "C1",
    "C2",
    "C3",
    "C4",
    "C5",
    "F1",
    "F2",
    "F3",
    "F4",
    "F6",
    "P1",
    "P2",
    "P4",
    "P6",
    "P13",
    "P14",
]

# C-type defined in Unidic-CSJ
C_TYPES_UNIDIC_CSJ = [
    "カ行変格",
    "サ行変格",
    "上一段-ア行",
    "上一段-カ行",
    "上一段-ガ行",
    "上一段-ザ行",
    "上一段-タ行",
    "上一段-ナ行",
    "上一段-ハ行",
    "上一段-バ行",
    "上一段-マ行",
    "上一段-ラ行",
    "下一段-ア行",
    "下一段-カ行",
    "下一段-ガ行",
    "下一段-サ行",
    "下一段-ザ行",
    "下一段-タ行",
    "下一段-ダ行",
    "下一段-ナ行",
    "下一段-ハ行",
    "下一段-バ行",
    "下一段-マ行",
    "下一段-ラ行",
    "五段-カ行",
    "五段-ガ行",
    "五段-サ行",
    "五段-タ行",
    "五段-ナ行",
    "五段-バ行",
    "五段-マ行",
    "五段-ラ行",
    "五段-ワア行",
    "助動詞-ジャ",
    "助動詞-タ",
    "助動詞-タイ",
    "助動詞-ダ",
    "助動詞-デス",
    "助動詞-ドス",
    "助動詞-ナイ",
    "助動詞-ナンダ",
    "助動詞-ヌ",
    "助動詞-ヒン",
    "助動詞-ヘン",
    "助動詞-マイ",
    "助動詞-マス",
    "助動詞-ヤ",
    "助動詞-ヤス",
    "助動詞-ヤン",
    "助動詞-ラシイ",
    "助動詞-レル",
    "助動詞-ンス",
    "形容詞",
    "文語カ行変格",
    "文語サ行変格",
    "文語ラ行変格",
    "文語上一段-マ行",
    "文語上一段-ワ行",
    "文語上二段-タ行",
    "文語上二段-バ行",
    "文語上二段-ラ行",
    "文語下二段-ア行",
    "文語下二段-カ行",
    "文語下二段-ガ行",
    "文語下二段-サ行",
    "文語下二段-タ行",
    "文語下二段-ダ行",
    "文語下二段-ナ行",
    "文語下二段-ハ行",
    "文語下二段-マ行",
    "文語下二段-ラ行",
    "文語下二段-ワ行",
    "文語助動詞-キ",
    "文語助動詞-ケリ",
    "文語助動詞-ゴトシ",
    "文語助動詞-ズ",
    "文語助動詞-タリ-完了",
    "文語助動詞-タリ-断定",
    "文語助動詞-ナリ-断定",
    "文語助動詞-ベシ",
    "文語助動詞-マジ",
    "文語助動詞-リ",
    "文語四段-カ行",
    "文語四段-ガ行",
    "文語四段-サ行",
    "文語四段-タ行",
    "文語四段-ハ行",
    "文語四段-バ行",
    "文語四段-マ行",
    "文語四段-ラ行",
    "文語形容詞-ク",
    "文語形容詞-シク",
    "無変化型",
    "特殊型",
]

# C-type defined in OpenJTalk dictionary (NAIST-jdic)
C_TYPES_OPEN_JTALK = [
    "カ変・クル",
    "カ変・来ル",
    "サ変・スル",
    "サ変・－スル",
    "サ変・－ズル",
    "ラ変",
    "一段",
    "一段・クレル",
    "一段・得ル",
    "上二・ダ行",
    "上二・ハ行",
    "下二・カ行",
    "下二・ガ行",
    "下二・タ行",
    "下二・ダ行",
    "下二・ハ行",
    "下二・マ行",
    "下二・得",
    "不変化型",
    "五段・カ行イ音便",
    "五段・カ行促音便",
    "五段・カ行促音便ユク",
    "五段・ガ行",
    "五段・サ行",
    "五段・タ行",
    "五段・ナ行",
    "五段・バ行",
    "五段・マ行",
    "五段・ラ行",
    "五段・ラ行アル",
    "五段・ラ行特殊",
    "五段・ワ行ウ音便",
    "五段・ワ行促音便",
    "四段・サ行",
    "四段・タ行",
    "四段・ハ行",
    "四段・バ行",
    "形容詞・アウオ段",
    "形容詞・イ段",
    "文語・キ",
    "文語・ケリ",
    "文語・ゴトシ",
    "文語・ナリ",
    "文語・ベシ",
    "文語・マジ",
    "文語・リ",
    "文語・ル",
    "特殊・ジャ",
    "特殊・タ",
    "特殊・タイ",
    "特殊・ダ",
    "特殊・デス",
    "特殊・ナイ",
    "特殊・ヌ",
    "特殊・マス",
    "特殊・ヤ",
]

# C-form defined in Unidic-CSJ
C_FORMS_UNIDIC_CSJ = [
    "ク語法",
    "仮定形-一般",
    "仮定形-融合",
    "命令形",
    "已然形-一般",
    "已然形-補助",
    "意志推量形",
    "未然形-サ",
    "未然形-セ",
    "未然形-一般",
    "未然形-撥音便",
    "未然形-補助",
    "終止形-一般",
    "終止形-促音便",
    "終止形-撥音便",
    "終止形-融合",
    "終止形-補助",
    "語幹-サ",
    "語幹-一般",
    "連体形-イ音便",
    "連体形-ウ音便",
    "連体形-一般",
    "連体形-促音便",
    "連体形-撥音便",
    "連体形-省略",
    "連体形-補助",
    "連用形-イ音便",
    "連用形-ウ音便",
    "連用形-ト",
    "連用形-ニ",
    "連用形-一般",
    "連用形-促音便",
    "連用形-撥音便",
    "連用形-省略",
    "連用形-融合",
    "連用形-補助",
]

# C-form defined in OpenJTalk dictionary (NAIST-jdic)
C_FORMS_OPEN_JTALK = [
    "ガル接続",
    "仮定形",
    "仮定縮約１",
    "仮定縮約２",
    "体言接続",
    "体言接続特殊",
    "体言接続特殊２",
    "命令ｅ",
    "命令ｉ",
    "命令ｒｏ",
    "命令ｙｏ",
    "基本形",
    "文語基本形",
    "未然ウ接続",
    "未然ヌ接続",
    "未然レル接続",
    "未然形",
    "未然特殊",
    "現代基本形",
    "連用ゴザイ接続",
    "連用タ接続",
    "連用テ接続",
    "連用デ接続",
    "連用ニ接続",
    "連用形",
    "音便基本形",
]

# Accent-type support Unidic-CSJ
ACCENT_TYPES_UNIDIC_CSJ = list(range(23))

# Accent-type support OpenJTalk
ACCENT_TYPES_OPEN_JTALK = list(range(50))

# Chain falg of OpenJTalk
CHAIN_FLAG_OPEN_JTALK = list(range(2))

# Word-type Table defined in Unidic-CSJ
WORD_TYPES_UNIDIC_CSJ = ["和", "漢", "外", "混", "固", "記号", "他"]

# Accent-mod-type defined in Unidic-CSJ
ACCENT_MOD_TYPES_UNIDIC_CSJ = ["M1@0", "M1@1", "M2@0", "M2@1", "M2@2", "M4@1"]

FEATURE_TABLES = {
    "unidic-csj": {
        "surface": None,  # surface will be loaded from vocab.pkl
        "mora": MORAS,
        "pos": POS_UNIDIC_CSJ,
        "c_type": C_TYPES_UNIDIC_CSJ,
        "c_form": C_FORMS_UNIDIC_CSJ,
        "accent_type": ACCENT_TYPES_UNIDIC_CSJ,
        "word_type": WORD_TYPES_UNIDIC_CSJ,
        "accent_con_type": ACCENT_CON_TYPES_UNIDIC_CSJ,
        "accent_mod_type": ACCENT_MOD_TYPES_UNIDIC_CSJ,
    },
    "open-jtalk": {
        "surface": None,  # surface will be loaded from vocab.pkl
        "mora": MORAS,
        "pos": POS_OPEN_JTALK,
        "c_type": C_TYPES_OPEN_JTALK,
        "c_form": C_FORMS_OPEN_JTALK,
        "accent_type": ACCENT_TYPES_OPEN_JTALK,
        "accent_con_type": ACCENT_CON_TYPES_OPEN_JTALK,
        "chain_flag": CHAIN_FLAG_OPEN_JTALK,
    },
}

RAW_FEATURE_KEYS = {
    "unidic-csj": [
        "surface",
        "pron",
        "pos",
        "c_type",
        "c_form",
        "accent_type",
        "word_type",
        "accent_con_type",
        "accent_mod_type",
    ],
    "open-jtalk": [
        "surface",
        "pron",
        "pos",
        "c_type",
        "c_form",
        "accent_type",
        "accent_con_type",
        "chain_flag",
    ],
}

_re_accent_con_f_type = re.compile(r"((形容詞|名詞|動詞)\%(F\d))")


def is_noun(pos_tag):
    return pos_tag.startswith("名詞")


def is_verb(pos_tag):
    return pos_tag.startswith("動詞")


def is_adjective(pos_tag):
    return pos_tag.startswith("形容詞")


def _pos_id_to_tag(pos_tag, unk_token):
    if is_noun(pos_tag):
        pos_key = "名詞"
    elif is_verb(pos_tag):
        pos_key = "動詞"
    elif is_adjective(pos_tag):
        pos_key = "形容詞"
    else:
        pos_key = unk_token

    return pos_key


def parse_accent_con_type(a_con_type, pos, unk_token="[UNK]"):
    if a_con_type is None:
        a_con_type = unk_token
    elif "%" in a_con_type:
        a_con_type = {
            con[1]: con[2] for con in _re_accent_con_f_type.findall(a_con_type)
        }
        pos_tag = _pos_id_to_tag(pos, unk_token)
        a_con_type = a_con_type.get(pos_tag, unk_token)

    return a_con_type



import argparse
import json
import random
import sys
from pathlib import Path

import torch
from marine.data.feature.feature_set import FeatureSet
from marine.data.util import load_dataset
from marine.logger import getLogger
from marine.models import (
    AttentionBasedLSTMDecoder,
    CRFDecoder,
    LinearDecoder,
    init_model,
)
from marine.utils.metrics import MultiTaskMetrics
from marine.utils.util import (
    convert_readable_labels,
    group_by_script_id,
    init_seed,
    log_scores,
    pack_inputs,
    pack_outputs,
    pad_incomplete_accent_logits,
    plot_batch_attention,
)
from omegaconf import OmegaConf
from tqdm import tqdm

logger = None


def get_parser():
    parser = argparse.ArgumentParser(
        description="Test model",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "checkpoint_dir", type=Path, help="Directory for checkpoint to test"
    )
    parser.add_argument(
        "--out_dir", "-o", type=Path, default=None, help="Directory test log"
    )
    parser.add_argument(
        "--checkpoint_filename",
        "-f",
        type=str,
        default="latest.pth",
        help="Model's file name to test",
    )
    parser.add_argument(
        "--data_dir", "-d", type=Path, default=None, help="Directory of dataset to test"
    )
    parser.add_argument(
        "--vocab_path", "-b", type=Path, default=None, help="Path of vocab file"
    )
    parser.add_argument("--n_jobs", type=int, default=8, help="Number of jobs")
    parser.add_argument(
        "--accent_status_represent_mode",
        "-m",
        type=str,
        choices=["binary", "high_low"],
        default="binary",
        help="Representation mode for accent status label",
    )
    parser.add_argument(
        "--random_seed",
        "-r",
        type=int,
        default=12345,
        help="Random seed for sampling",
    )
    parser.add_argument(
        "--verbose",
        "-v",
        type=int,
        default=50,
        help="Logging level",
    )
    return parser


def test_model(
    model,
    checkpoint_dir,
    checkpoint_file,
    tasks,
    dataloader,
    config,
    feature_set,
    tensorboard_writer=None,
    logger=None,
    device="cpu",
):
    model_path = checkpoint_dir / checkpoint_file
    states = torch.load(model_path)

    phase = "test"
    fig_logging_targets = random.choices(range(config.data.batch_size), k=10)

    logger.info(f"Load checkpoint from {model_path} ({states['epoch']}th epoch)")
    model.load_state_dict(states["state_dict"])

    dataloader = dataloader[phase]

    if "accent_status" in tasks:
        has_att_based_model = isinstance(
            model.decoders["accent_status"], AttentionBasedLSTMDecoder
        )
    else:
        has_att_based_model = False

    # for logging
    metrics = MultiTaskMetrics(
        phase,
        config.data.output_sizes,
        accent_represent_mode=config.data.represent_mode,
        require_ap_level_f1_score=has_att_based_model,
        device=device,
    )

    total_logs = {task: {} for task in tasks}

    for batch_index, (inputs, outputs, _, script_ids) in enumerate(
        tqdm(dataloader, desc=f"{phase}: ", leave=False)
    ):
        # pack inputs to device
        inputs = pack_inputs(inputs, config.data.input_keys, device)
        outputs = pack_outputs(outputs, device)

        prev_decoder_output = {}

        for task_index, task in enumerate(tasks):
            output, output_mask = outputs[task]["label"], outputs[task]["mask"]
            decoder_outputs = model(task, **inputs)

            # predict
            if isinstance(model.decoders[task], CRFDecoder):
                _, crf_logits = decoder_outputs
                logits = crf_logits
            elif isinstance(model.decoders[task], LinearDecoder):
                logits = decoder_outputs
            elif isinstance(model.decoders[task], AttentionBasedLSTMDecoder):
                logits, attentions, ap_lengths = decoder_outputs

                # plot attention when first batch on test
                if tensorboard_writer and batch_index == 0 and task == "accent_status":
                    plot_batch_attention(
                        inputs,
                        logits,
                        ap_lengths,
                        attentions,
                        feature_set,
                        plot_targets=fig_logging_targets,
                        tensorboard_writer=tensorboard_writer,
                        phase=phase,
                        epoch=0,
                        script_ids=script_ids,
                    )

                # resize logits
                logits = pad_incomplete_accent_logits(logits, output_mask)

            # logits: (B, T, dim) -> (B, T)
            predicts = torch.argmax(logits, dim=2)

            # Log predicts
            total_logs[task].update(
                convert_readable_labels(predicts, output, output_mask, script_ids)
            )

            # Update metrics
            # for ap-based seq
            if task == "accent_status" and has_att_based_model:
                metrics.update(
                    task,
                    predicts,
                    output,
                    output_mask,
                    predicted_accent_phrase_boundaries=inputs["prev_decoder_outputs"][
                        "accent_phrase_boundary"
                    ],
                    target_accent_phrase_boundaries=outputs["accent_phrase_boundary"][
                        "label"
                    ],
                    mora_seq_masks=outputs["accent_phrase_boundary"]["mask"],
                )

            # for mora-based seq
            else:
                metrics.update(task, predicts, output, output_mask)

            if task_index < len(tasks) - 1:
                prev_decoder_output[task] = predicts
                inputs["prev_decoder_outputs"][task] = predicts

    # Group by script_id
    total_logs = group_by_script_id(total_logs)

    # Logging scores
    log_scores(
        phase=phase,
        epoch=0,
        tasks=tasks,
        metrics=metrics,
        logs=total_logs,
        loss=None,
        tensorboard_writer=tensorboard_writer,
    )

    return total_logs


def entry(argv=sys.argv):
    global logger
    args = get_parser().parse_args(argv[1:])
    logger = getLogger(args.verbose)
    logger.debug(f"Loaded parameters: {args}")

    init_seed(args.random_seed)

    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    checkpoint_dir = args.checkpoint_dir

    if not checkpoint_dir.exists():
        raise FileNotFoundError("Checkpoint dir not found")

    checkpoint_config_path = checkpoint_dir / "config.yaml"

    if not checkpoint_config_path.exists():
        raise FileNotFoundError("config file not found")

    checkpoint_config = OmegaConf.load(checkpoint_config_path)

    logger.info("Loaded config")
    logger.info(checkpoint_config)

    if args.data_dir:
        checkpoint_config.data.data_dir = str(args.data_dir)

    checkpoint_config.data.num_workers = args.n_jobs
    checkpoint_config.data.represent_mode = args.accent_status_represent_mode

    dataloader = load_dataset(checkpoint_config, phases=["test"])
    tasks = checkpoint_config.data.output_keys

    if args.out_dir:
        log_dir = Path(args.out_dir)
    else:
        log_dir = Path("logs") / checkpoint_dir.name

    if not log_dir.exists():
        log_dir.mkdir(parents=True)

    log_path = log_dir / f"{checkpoint_dir.name}_test_log.json"

    # init feature set
    if args.vocab_path:
        if args.vocab_path.exists():
            feature_set = FeatureSet(
                args.vocab_path,
                feature_table_key=checkpoint_config.data.feature_table_key,
                feature_keys=checkpoint_config.data.input_keys,
            )
        else:
            raise FileNotFoundError(f"Not found vocab file: {args.vocab_path}")
    else:
        raise FileNotFoundError(
            "Please specify vocab file path in args or config/model/*.yaml"
        )

    # Init test model
    model = init_model(tasks, checkpoint_config, feature_set, device)
    logs = test_model(
        model,
        checkpoint_dir,
        args.checkpoint_filename,
        tasks,
        dataloader,
        checkpoint_config,
        feature_set,
        logger=logger,
        device=device,
    )

    # save log
    with open(log_path, "w") as file:
        json.dump(logs, file, ensure_ascii=False, indent=4, separators=(",", ": "))


if __name__ == "__main__":
    sys.exit(entry())



from torch.utils.data import Dataset


class AccentDataset(Dataset):
    def __init__(self, data):
        self.data = data

    def __len__(self):
        return len(self.data["labels"])

    def __getitem__(self, index):
        item = {
            "features": self.data["features"][index],
            "labels": self.data["labels"][index],
        }

        if "ids" in self.data.keys():
            item["ids"] = self.data["ids"][index]

        return item



# coding: utf-8

from torch import nn


class BaseModel(nn.Module):
    def __init__(self, embedding, encoders, decoders):
        super().__init__()
        self.embedding = embedding
        self.encoders = nn.ModuleDict(encoders)
        self.decoders = nn.ModuleDict(decoders)

    def forward(
        self,
        task,
        embedding_features,
        lengths,
        mask,
        prev_decoder_outputs=None,
        decoder_targets=None,
    ):
        embeddings = self.embedding(**embedding_features)
        encoder_outputs = self.encoders[task](embeddings, lengths)
        decoder_outputs = self.decoders[task](
            encoder_outputs, mask, prev_decoder_outputs, decoder_targets
        )

        return decoder_outputs



# coding: utf-8

import re

_re_has_longvowel = None
_re_hiragana = None
_re_katakana = None
_re_kanji = None
_re_letter = None
_re_symbol = None


def is_hiragana(surface):
    global _re_hiragana
    if _re_hiragana is None:
        _re_hiragana = re.compile(r"^[ぁ-ん]+$")
    return _re_hiragana.match(surface) is not None


def is_katakana(surface):
    global _re_katakana
    if _re_katakana is None:
        _re_katakana = re.compile(r"^[ァ-ヴヶ]+$")
    return _re_katakana.match(surface) is not None


def is_kanji(surface):
    global _re_kanji
    if _re_kanji is None:
        _re_kanji = re.compile(r"^[一-龠]+$")
    return _re_kanji.match(surface) is not None


def is_letter(surface):
    global _re_letter
    if _re_letter is None:
        _re_letter = re.compile(r"^[a-zA-Z]+$")
    return _re_letter.match(surface) is not None


def is_symbol(surface):
    global _re_symbol
    if _re_symbol is None:
        _re_symbol = re.compile(r"^[〆々ー,.?!]+$")
    return _re_symbol.match(surface) is not None


def has_longvowel(text):
    global _re_has_longvowel
    if _re_has_longvowel is None:
        _re_has_longvowel = re.compile(r"(aa|ii|uu|ee|oo)$")
    return _re_has_longvowel.search(text) is not None



# flake8: noqa: B023

import argparse
import sys
from concurrent.futures import ProcessPoolExecutor
from multiprocessing import cpu_count
from pathlib import Path

import numpy as np
from joblib import dump, load
from marine.data.feature.feature_set import FeatureSet
from marine.logger import getLogger
from marine.utils.g2p_util import pron2mora
from marine.utils.util import load_json_corpus, split_corpus
from tqdm import tqdm

logger = None

AP_BOUNDARY_LABEL = 1


def get_parser():
    parser = argparse.ArgumentParser(
        description="Convert Special format txt format data to json file",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument("corpus_path", type=Path, help="Path or directory for corpus")
    parser.add_argument(
        "feature_path", type=Path, help="Path or directory for feature file"
    )
    parser.add_argument("vocab_path", type=Path, help="Vocab file path")
    parser.add_argument("out_dir", type=Path, help="Output directory")
    parser.add_argument("--n_jobs", type=int, default=8, help="Number of jobs")
    parser.add_argument(
        "--feature_table_key",
        "-f",
        type=str,
        choices=["unidic-csj", "open-jtalk"],
        default="unidic-csj",
        help="Sequence level for accent status label",
    )
    parser.add_argument(
        "--max_size",
        "-m",
        type=int,
        default=-1,
        help="""Maximum number of scripts to convert feature
        (-m < 0 = use all scripts)""",
    )
    parser.add_argument(
        "--accent_status_seq_level",
        "-s",
        type=str,
        choices=["ap", "mora"],
        default="ap",
        help="Sequence level for accent status label",
    )
    parser.add_argument(
        "--test_size",
        "-t",
        type=int,
        default=-1,
        help="""Specific size of samples for val and test
        (-t < 0 = 5%% and / 5%% for val and test respectively)""",
    )
    parser.add_argument(
        "--skip_corpus_split",
        action="store_true",
        help="Whether skip corpus splitting",
    )
    parser.add_argument(
        "--single_corpus_key",
        "-k",
        type=str,
        choices=["train", "val", "test"],
        default="train",
        help="""Dataset key to export
        (this option will be ignored if --skip_corpus_split is not activated)""",
    )
    parser.add_argument(
        "--target_id_dir",
        type=Path,
        default=None,
        help="Directory of id files to reproduce specific dataset",
    )
    parser.add_argument(
        "--random_seed",
        "-r",
        type=int,
        default=12345,
        help="Random seed for sampling",
    )
    parser.add_argument(
        "--verbose",
        "-v",
        type=int,
        default=50,
        help="Logging level",
    )
    return parser


def insert_punctuation_by_extracted_features(
    mora_seq_w_puncts,
    mora_seq_wo_puncts,
    labels,
    punctuation_ids,
    accent_status_seq_level,
):
    indexs = np.where(np.in1d(mora_seq_w_puncts, punctuation_ids))[0]

    for index in indexs:
        mora_seq_wo_puncts = np.insert(
            mora_seq_wo_puncts, index, int(mora_seq_w_puncts[index])
        )
        for label_key in labels.keys():
            if label_key == "accent_status" and accent_status_seq_level == "ap":
                # the AP-based AN label represents the position of AN in a AP
                # i.e., Don't need consider the global indexå
                # Also, the accent nucleus never located at punctuation
                continue

            # the label where punctuation are must be 0
            labels[label_key].insert(index, 0)
            assert len(mora_seq_wo_puncts) == len(
                labels[label_key]
            ), f"Length of inserted result is not matched: {len(mora_seq_wo_puncts)} != {len(labels[label_key])} ({label_key})"

    assert np.array_equal(
        mora_seq_w_puncts, mora_seq_wo_puncts
    ), f"The inserted mora seq not same with original mora seq: {mora_seq_w_puncts} != {mora_seq_wo_puncts}"

    return labels


def _process(
    nodes,
    feature_set,
    accent_status_seq_level,
    script_id,
    surface,
    pron,
    **original_labels,
):
    if len(nodes) == 0:
        if logger is not None:
            logger.debug(f"Empty nodes: {script_id}")
        return None

    # get punctuation id
    punctuation_ids = feature_set.get_punctuation_ids()

    # check accent seq level
    required_ap_accent = (
        "accent" in original_labels.keys() and accent_status_seq_level == "ap"
    )

    # init labels, features
    labels = {key: [] for key in original_labels.keys()}
    _original_labels = {
        key: [int(v) for v in value.split(",")]
        for key, value in original_labels.items()
    }

    # convert nodes to feature seqs
    feature = feature_set.convert_nodes_to_feature(nodes)

    # convert original pron in annotation to id seqs
    original_mora = pron2mora(pron)
    expected_ids = feature_set.convert_feature_to_id("mora", original_mora)

    # verify the features is available
    # i.e., is the prounnounces is same w/o punctuation
    punct_removed_extracted_mora = feature["mora"][
        np.in1d(feature["mora"], punctuation_ids, invert=True)
    ]
    punct_removed_expected_mora = expected_ids[
        np.in1d(expected_ids, punctuation_ids, invert=True)
    ]

    if np.array_equal(punct_removed_extracted_mora, punct_removed_expected_mora):
        if len(feature["mora"]) != len(expected_ids):
            _original_labels = insert_punctuation_by_extracted_features(
                feature["mora"],
                expected_ids,
                _original_labels,
                punctuation_ids,
                accent_status_seq_level,
            )

        for key in labels.keys():
            # verify for ap-based AN labels with AP labels
            if (
                required_ap_accent
                and key == "accent_status"
                and "accent_phrase_boundary" in labels.keys()
            ):
                num_boundary = (
                    np.count_nonzero(
                        _original_labels["accent_phrase_boundary"] == AP_BOUNDARY_LABEL
                    )
                    + 1
                )
                accents = _original_labels["accent_status"]
                assert (
                    len(accents) == num_boundary
                ), "Unmatched length of sequnce between ac and ap"
                labels["accent_status"] = np.array(
                    _original_labels["accent_status"], dtype=np.uint8
                )
            else:
                labels[key] = np.array(_original_labels[key], dtype=np.uint8)
    else:
        if logger is not None:
            logger.debug(
                (
                    f"Wrong mora [{script_id}]:"
                    f" {''.join(feature_set.convert_id_to_feature('mora', punct_removed_expected_mora))}"
                    f" != {''.join(feature_set.convert_id_to_feature('mora', punct_removed_extracted_mora))}"
                )
            )
        return None

    return script_id, feature, labels


def _load_target_ids(target_id_dir):
    id_groups = {}

    logger.info(f"Load ids from existing dataset ({target_id_dir})")
    for path in target_id_dir.glob("*/ids.pkl"):
        dataset_key = str(path.parent.name)
        id_groups[dataset_key] = set(load(path))
        logger.info(
            f"Loaded {len(id_groups[dataset_key]):,} ids for {dataset_key} in {target_id_dir}"
        )

    return id_groups


def _sort_corpus_by_script_id(corpus):
    return list(sorted(corpus, key=lambda x: x[0]))


def _remove_unavailable_script(corpus):
    return list(filter(lambda x: x is not None, corpus))


def _split_corpus_by_ids(corpus, id_groups):
    _corpus = {}

    for key, target_ids in id_groups.items():
        scripts = list(filter(lambda x: x[0] in target_ids, corpus))
        assert len(scripts) == len(
            target_ids
        ), f"Not enough number of scripts for {key}: {len(scripts):,} != {len(target_ids):,}"
        _corpus[key] = scripts

    # When specified the ID group including in val, test
    if set(["val", "test"]) == set(_corpus.keys()):
        _corpus["train"] = list(
            filter(
                lambda x: x[0] not in id_groups["val"]
                and x[0] not in id_groups["test"],
                corpus,
            )
        )
    # When the ID group is specified fully (i.e., train, val and test)
    elif set(["train", "val", "test"]) == set(_corpus.keys()):
        pass
    else:
        raise NotImplementedError(
            f"Not supported ID group specification: {_corpus.keys()}"
        )

    return _corpus


def entry(argv=sys.argv):
    global logger

    args = get_parser().parse_args(argv[1:])
    logger = getLogger(args.verbose)
    logger.debug(args)

    # Init feature-set
    feature_set = FeatureSet(args.vocab_path, feature_table_key=args.feature_table_key)

    if not args.out_dir.exists():
        args.out_dir.mkdir(parents=True)

    # Load corpus
    corpus = load_json_corpus(args.corpus_path)
    features = load_json_corpus(args.feature_path)

    assert len(corpus) == len(
        features
    ), "Not match script size between corpus and feature files"
    assert [script["script_id"] for script in corpus] == [
        script["script_id"] for script in features
    ], "Not match script ids between corpus and feature files"

    if args.max_size > 0:
        assert (
            len(corpus) > args.max_size
        ), f"Not enough number of script: {len(corpus)} < {args.max_size}"
        corpus = corpus[: args.max_size]
        features = features[: args.max_size]

    # Process
    n_jobs = min(cpu_count(), args.n_jobs)

    if n_jobs > 1:
        logger.info(f"Processing {len(corpus):,} scripts with {n_jobs} jobs")
        with ProcessPoolExecutor(n_jobs) as executor:
            futures = [
                executor.submit(
                    _process,
                    feature["nodes"],
                    feature_set,
                    args.accent_status_seq_level,
                    **script,
                )
                for script, feature in zip(corpus, features)
            ]
            corpus = [
                future.result()
                for future in tqdm(
                    futures, desc="Convert corpus to feature", leave=False
                )
            ]
    else:
        logger.info(f"Processing {len(corpus):,} scripts in a single thread")
        corpus = [
            _process(
                feature["nodes"], feature_set, args.accent_status_seq_level, **script
            )
            for script, feature in tqdm(
                zip(corpus, features), desc="Convert corpus to feature", leave=False
            )
        ]

    # Cleaing
    corpus = _sort_corpus_by_script_id(_remove_unavailable_script(corpus))
    logger.info(f"Cleaned corpus: {len(corpus):,}")

    if args.target_id_dir is None:
        if args.skip_corpus_split:
            corpus = {key: corpus for key in [args.single_corpus_key]}
        else:
            corpus = split_corpus(
                corpus, random_state=args.random_seed, absolute_test_size=args.test_size
            )
    else:
        id_groups = _load_target_ids(args.target_id_dir)
        corpus = _split_corpus_by_ids(corpus, id_groups)

    # Output results
    for key in corpus:
        phase_dir = args.out_dir / key
        phase_dir.mkdir(parents=True, exist_ok=True)

        _corpus = _sort_corpus_by_script_id(corpus[key])

        ids, features, labels = zip(*_corpus)
        logger.info(f"Saving {key} file with {len(labels):,} corpus in {args.out_dir}")

        dump(list(ids), phase_dir / "ids.pkl", compress=True)
        dump(list(features), phase_dir / "features.pkl", compress=True)
        dump(list(labels), phase_dir / "labels.pkl", compress=True)


if __name__ == "__main__":
    sys.exit(entry())



# Copyright 2019 Allen Institute for AI
#  Apache 2.0  (http://www.apache.org/licenses/LICENSE-2.0)

import random
from logging import getLogger
from typing import Any

import torch
from marine.modules.crf_tagger import logsumexp, viterbi_decode
from numpy.testing import assert_almost_equal, assert_equal

logger = getLogger("test")


def test_logsumexp():
    # First a simple example where we add probabilities in log space.
    tensor = torch.FloatTensor([[0.4, 0.1, 0.2]])
    log_tensor = tensor.log()
    log_summed = logsumexp(log_tensor, dim=-1, keepdim=False)
    assert_almost_equal(log_summed.exp().data.numpy(), [0.7])
    log_summed = logsumexp(log_tensor, dim=-1, keepdim=True)
    assert_almost_equal(log_summed.exp().data.numpy(), [[0.7]])

    # Then some more atypical examples, and making sure this will work with how we handle
    # log masks.
    tensor = torch.FloatTensor([[float("-inf"), 20.0]])
    assert_almost_equal(logsumexp(tensor).data.numpy(), [20.0])
    tensor = torch.FloatTensor([[-200.0, 20.0]])
    assert_almost_equal(logsumexp(tensor).data.numpy(), [20.0])
    tensor = torch.FloatTensor([[20.0, 20.0], [-200.0, 200.0]])
    assert_almost_equal(logsumexp(tensor, dim=0).data.numpy(), [20.0, 200.0])


def test_viterbi_decode():
    # Test Viterbi decoding is equal to greedy decoding with no pairwise potentials.
    sequence_logits = torch.nn.functional.softmax(torch.rand([5, 9]), dim=-1)
    transition_matrix = torch.zeros([9, 9])
    indices, _ = viterbi_decode(sequence_logits.data, transition_matrix)
    _, argmax_indices = torch.max(sequence_logits, 1)
    assert indices == argmax_indices.data.squeeze().tolist()

    # Test Viterbi decoding works with start and end transitions
    sequence_logits = torch.nn.functional.softmax(torch.rand([5, 9]), dim=-1)
    transition_matrix = torch.zeros([9, 9])
    allowed_start_transitions = torch.zeros([9])
    # Force start tag to be an 8
    allowed_start_transitions[:8] = float("-inf")
    allowed_end_transitions = torch.zeros([9])
    # Force end tag to be a 0
    allowed_end_transitions[1:] = float("-inf")
    indices, _ = viterbi_decode(
        sequence_logits.data,
        transition_matrix,
        allowed_end_transitions=allowed_end_transitions,
        allowed_start_transitions=allowed_start_transitions,
    )
    assert indices[0] == 8
    assert indices[-1] == 0

    # Test that pairwise potentials affect the sequence correctly and that
    # viterbi_decode can handle -inf values.
    sequence_logits = torch.FloatTensor(
        [
            [0, 0, 0, 3, 5],
            [0, 0, 0, 3, 4],
            [0, 0, 0, 3, 4],
            [0, 0, 0, 3, 4],
            [0, 0, 0, 3, 4],
            [0, 0, 0, 3, 4],
        ]
    )
    # The same tags shouldn't appear sequentially.
    transition_matrix = torch.zeros([5, 5])
    for i in range(5):
        transition_matrix[i, i] = float("-inf")
    indices, _ = viterbi_decode(sequence_logits, transition_matrix)
    assert indices == [4, 3, 4, 3, 4, 3]

    # Test that unbalanced pairwise potentials break ties
    # between paths with equal unary potentials.
    sequence_logits = torch.FloatTensor(
        [
            [0, 0, 0, 4, 4],
            [0, 0, 0, 4, 4],
            [0, 0, 0, 4, 4],
            [0, 0, 0, 4, 4],
            [0, 0, 0, 4, 4],
            [0, 0, 0, 4, 4],
        ]
    )
    # The 5th tag has a penalty for appearing sequentially
    # or for transitioning to the 4th tag, making the best
    # path uniquely to take the 4th tag only.
    transition_matrix = torch.zeros([5, 5])
    transition_matrix[4, 4] = -10
    transition_matrix[4, 3] = -10
    transition_matrix[3, 4] = -10
    indices, _ = viterbi_decode(sequence_logits, transition_matrix)
    assert indices == [3, 3, 3, 3, 3, 3]

    sequence_logits = torch.FloatTensor([[1, 0, 0, 4], [1, 0, 6, 2], [0, 3, 0, 4]])
    # Best path would normally be [3, 2, 3] but we add a
    # potential from 2 -> 1, making [3, 2, 1] the best path.
    transition_matrix = torch.zeros([4, 4])
    transition_matrix[0, 0] = 1
    transition_matrix[2, 1] = 5
    indices, value = viterbi_decode(sequence_logits, transition_matrix)
    assert indices == [3, 2, 1]
    assert value.numpy() == 18

    # Test that providing evidence results in paths containing specified tags.
    sequence_logits = torch.FloatTensor(
        [
            [0, 0, 0, 7, 7],
            [0, 0, 0, 7, 7],
            [0, 0, 0, 7, 7],
            [0, 0, 0, 7, 7],
            [0, 0, 0, 7, 7],
            [0, 0, 0, 7, 7],
        ]
    )
    # The 5th tag has a penalty for appearing sequentially
    # or for transitioning to the 4th tag, making the best
    # path to take the 4th tag for every label.
    transition_matrix = torch.zeros([5, 5])
    transition_matrix[4, 4] = -10
    transition_matrix[4, 3] = -2
    transition_matrix[3, 4] = -2
    # The 1st, 4th and 5th sequence elements are observed - they should be
    # equal to 2, 0 and 4. The last tag should be equal to 3, because although
    # the penalty for transitioning to the 4th tag is -2, the unary potential
    # is 7, which is greater than the combination for any of the other labels.
    observations = [2, -1, -1, 0, 4, -1]
    indices, _ = viterbi_decode(sequence_logits, transition_matrix, observations)
    assert indices == [2, 3, 3, 0, 4, 3]


def test_viterbi_decode_top_k():
    # Test cases taken from: https://gist.github.com/PetrochukM/afaa3613a99a8e7213d2efdd02ae4762

    # Test Viterbi decoding is equal to greedy decoding with no pairwise potentials.
    sequence_logits = torch.autograd.Variable(torch.rand([5, 9]))
    transition_matrix = torch.zeros([9, 9])

    indices, _ = viterbi_decode(sequence_logits.data, transition_matrix, top_k=5)

    _, argmax_indices = torch.max(sequence_logits, 1)
    assert indices[0] == argmax_indices.data.squeeze().tolist()

    # Test that pairwise potentials effect the sequence correctly and that
    # viterbi_decode can handle -inf values.
    sequence_logits = torch.FloatTensor(
        [
            [0, 0, 0, 3, 4],
            [0, 0, 0, 3, 4],
            [0, 0, 0, 3, 4],
            [0, 0, 0, 3, 4],
            [0, 0, 0, 3, 4],
            [0, 0, 0, 3, 4],
        ]
    )
    # The same tags shouldn't appear sequentially.
    transition_matrix = torch.zeros([5, 5])
    for i in range(5):
        transition_matrix[i, i] = float("-inf")
    indices, _ = viterbi_decode(sequence_logits, transition_matrix, top_k=5)
    assert indices[0] == [3, 4, 3, 4, 3, 4]

    # Test that unbalanced pairwise potentials break ties
    # between paths with equal unary potentials.
    sequence_logits = torch.FloatTensor(
        [
            [0, 0, 0, 4, 4],
            [0, 0, 0, 4, 4],
            [0, 0, 0, 4, 4],
            [0, 0, 0, 4, 4],
            [0, 0, 0, 4, 4],
            [0, 0, 0, 4, 0],
        ]
    )
    # The 5th tag has a penalty for appearing sequentially
    # or for transitioning to the 4th tag, making the best
    # path uniquely to take the 4th tag only.
    transition_matrix = torch.zeros([5, 5])
    transition_matrix[4, 4] = -10
    transition_matrix[4, 3] = -10
    indices, _ = viterbi_decode(sequence_logits, transition_matrix, top_k=5)
    assert indices[0] == [3, 3, 3, 3, 3, 3]

    sequence_logits = torch.FloatTensor([[1, 0, 0, 4], [1, 0, 6, 2], [0, 3, 0, 4]])
    # Best path would normally be [3, 2, 3] but we add a
    # potential from 2 -> 1, making [3, 2, 1] the best path.
    transition_matrix = torch.zeros([4, 4])
    transition_matrix[0, 0] = 1
    transition_matrix[2, 1] = 5
    indices, value = viterbi_decode(sequence_logits, transition_matrix, top_k=5)
    assert indices[0] == [3, 2, 1]
    assert value[0] == 18

    def _brute_decode(
        tag_sequence: torch.Tensor, transition_matrix: torch.Tensor, top_k: int = 5
    ) -> Any:
        """
        Top-k decoder that uses brute search
        instead of the Viterbi Decode dynamic programing algorithm
        """
        # Create all possible sequences
        sequences = [[]]  # type: ignore

        for i in range(len(tag_sequence)):
            new_sequences = []  # type: ignore
            for j in range(len(tag_sequence[i])):
                for sequence in sequences:
                    new_sequences.append(sequence[:] + [j])
            sequences = new_sequences

        # Score
        scored_sequences = []  # type: ignore
        for sequence in sequences:
            emission_score = sum(tag_sequence[i, j] for i, j in enumerate(sequence))
            transition_score = sum(
                transition_matrix[sequence[i - 1], sequence[i]]
                for i in range(1, len(sequence))
            )
            score = emission_score + transition_score
            scored_sequences.append((score, sequence))

        # Get the top k scores / paths
        top_k_sequences = sorted(scored_sequences, key=lambda r: r[0], reverse=True)[
            :top_k
        ]
        scores, paths = zip(*top_k_sequences)

        return paths, scores  # type: ignore

    def _sanitize(x: Any) -> Any:
        """
        Sanitize turns PyTorch and Numpy types into basic Python types
        """
        if isinstance(x, (str, float, int, bool)):
            return x
        elif isinstance(x, torch.Tensor):
            return x.cpu().tolist()
        elif isinstance(x, (list, tuple, set)):
            return [_sanitize(x_i) for x_i in x]
        else:
            raise ValueError(f"Cannot sanitize {x} of type {type(x)}. ")

    # Run 100 randomly generated parameters and compare the outputs.
    for _ in range(100):
        num_tags = random.randint(1, 5)
        seq_len = random.randint(1, 5)
        k = random.randint(1, 5)
        sequence_logits = torch.rand([seq_len, num_tags])
        transition_matrix = torch.rand([num_tags, num_tags])
        viterbi_paths_v1, viterbi_scores_v1 = viterbi_decode(
            sequence_logits, transition_matrix, top_k=k
        )
        viterbi_path_brute, viterbi_score_brute = _brute_decode(
            sequence_logits, transition_matrix, top_k=k
        )
        assert_almost_equal(
            list(viterbi_score_brute), viterbi_scores_v1.tolist(), decimal=3
        )
        assert_equal(_sanitize(viterbi_paths_v1), viterbi_path_brute)



import re
from csv import reader

import numpy as np
from marine.utils.g2p_util import pron2mora
from marine.utils.g2p_util.g2p import ACCENT_REPRESENT_FUNC_TABLE

FEATURE_PARSE_SYMBOL = "/"
FEATURE_NODE_SPLIT_SYMBOL = "_"
FEATURE_MORA_SPLIT_SYMBOL = ","
FEATURE_ACCENT_SYMBOL = "@"

FEATURE_SYOMBOL_REMOVER = str.maketrans(
    {
        FEATURE_NODE_SPLIT_SYMBOL: "",
        FEATURE_MORA_SPLIT_SYMBOL: "",
        FEATURE_ACCENT_SYMBOL: "",
    }
)

PADDING_VALUE_FOR_LABEL = 1


def _make_align_array(surfaces):
    aligns = []
    index = 0

    while index < len(surfaces):
        current_surface = surfaces[index]
        blank_len = len(current_surface) - 1

        boundary = [index + 1]
        blank = [0] * blank_len if blank_len >= 1 else []

        aligns = aligns + boundary + blank

        index += 1

    return aligns


def _is_available_match(aligns, head, tail):
    return (head >= 0 and aligns[head] > 0) and (
        (tail == len(aligns)) or (tail < len(aligns) and aligns[tail] != 0)
    )


def _search_mark(padded_aligns):
    while len(padded_aligns) > 0:
        mark = padded_aligns.pop(-1)
        if mark > 0:
            return mark
    return -1


def aligns2mask(aligns, head, tail):
    if _is_available_match(aligns, head, tail):
        start = aligns[head] - 1
        end = _search_mark(aligns[head:tail])
        return (start, end)
    else:
        return None


def convert_feature_to_value(target, pron, label):
    if target == "accent_status":
        moras = pron2mora(pron)
        value = {}

        for accent_represent_mode in ACCENT_REPRESENT_FUNC_TABLE.keys():
            _, represented_accent = pron2mora(moras, label, accent_represent_mode)
            value[accent_represent_mode] = represented_accent
    else:
        moras = pron2mora(pron)
        value = len(moras) * [0]

        if label > 1:
            value[label - 1] = 1

    return moras, value


def load_postprocess_vocab(vocab_dir, tasks):
    vocab = {key: {} for key in tasks}

    for dict_dir in vocab_dir.iterdir():
        target = dict_dir.name

        if target == "vocab.pkl":
            continue

        assert target in tasks

        for dict_path in dict_dir.glob("*.tsv"):
            with dict_path.open() as dict_file:
                table = reader(dict_file, delimiter="\t")

                for pattern, value in table:
                    regex = re.compile(pattern)
                    surface, feature = value.split(FEATURE_PARSE_SYMBOL)
                    surfaces = surface.split(FEATURE_NODE_SPLIT_SYMBOL)
                    pron = feature.translate(FEATURE_SYOMBOL_REMOVER)

                    features = [
                        [mora for mora in moras.split(FEATURE_MORA_SPLIT_SYMBOL)]
                        for moras in feature.split(FEATURE_NODE_SPLIT_SYMBOL)
                    ]

                    assert len(surfaces) == len(
                        features
                    ), f"Wrong length entry : ({surfaces} != {features})"

                    labels = [
                        (0 if node_index == 0 else len(features[node_index - 1]))
                        + mora_index
                        + 1
                        for node_index, moras in enumerate(features)
                        for mora_index, mora in enumerate(moras)
                        if mora.endswith(FEATURE_ACCENT_SYMBOL)
                    ]

                    if labels:
                        # only use first appeared symbol
                        label = labels[0]
                    else:
                        label = -1

                    moras, values = convert_feature_to_value(target, pron, label)
                    vocab[target][pattern] = (regex, moras, values)

    return vocab


def apply_postprocess_dict(
    task,
    nodes,
    labels,
    moras,
    boundary,
    postprocess_targets,
    postprocess_vocab,
    accent_represent_mode="binary",
):
    surfaces = [node["surface"] for node in nodes]
    surface = "".join(surfaces)

    targets = postprocess_targets.findall(surface)

    if targets:
        aligns = _make_align_array(surfaces)

        for target in targets:
            regex, pron, values = postprocess_vocab[target]

            for match in regex.finditer(surface):
                head, tail = match.span()

                node_mask = aligns2mask(aligns, head, tail)

                if node_mask:
                    # get mora-based boundary's position
                    boundary_indexs = (
                        [0] + list(np.where(boundary > 0)[0]) + [len(moras)]
                    )
                    node_start, node_end = node_mask

                    mora_mask = slice(
                        boundary_indexs[node_start],
                        boundary_indexs[node_end],
                    )

                    if moras[mora_mask] == pron:
                        if task == "accent_status":
                            value = values[accent_represent_mode]
                        else:
                            value = values

                        labels[mora_mask] = value

    return labels



from torch.nn.modules.loss import _Loss


class LogLikelhood(_Loss):
    def __init__(self, log_likehood_func):
        super(LogLikelhood, self).__init__()
        self.log_likehood_func = log_likehood_func

    def forward(self, classified, label, mask=None):
        batch_size = label.size(0)
        log_likelihood = self.log_likehood_func(classified, label, mask)

        return -log_likelihood / batch_size



import json
from pathlib import Path
from typing import Dict

import numpy as np
import pytest
import torch
from hydra import compose, initialize_config_dir
from hydra.core.global_hydra import GlobalHydra
from marine.data.feature.feature_set import FeatureSet
from marine.data.feature.feature_table import (
    is_adjective,
    is_noun,
    is_verb,
    parse_accent_con_type,
)
from marine.models.util import init_model
from marine.utils.util import (
    _calculate_multiple_task_scores,
    _convert_ap_based_accent_to_mora_based_accent,
    convert_open_jtalk_format_label,
    expand_word_label_to_mora,
)
from numpy.testing import assert_almost_equal
from omegaconf import DictConfig
from pkg_resources import resource_filename

BASE_DIR = Path(resource_filename("marine", ""))


@pytest.fixture
def default_config() -> DictConfig:
    config_path = BASE_DIR / "bin" / "conf" / "train" / "config.yaml"

    # initialize config
    with initialize_config_dir(config_dir=str(config_path.parent)):
        config = compose(config_name=config_path.name)
    GlobalHydra.instance().clear()

    return config


@pytest.fixture
def default_vocab_path() -> Path:
    return (
        BASE_DIR.parent
        / "recipe"
        / "common"
        / "database"
        / "20220912_jsut_vocab_min_2"
        / "vocab.pkl"
    )


@pytest.fixture
def test_log_sample() -> Dict:
    logs = None
    sample_path = BASE_DIR.parent / "tests" / "samples" / "test_log_sample.json"
    with open(sample_path, "r") as file:
        logs = json.load(file)
    return logs


def test_label_expanding():
    """Test function to convert mora-based IP sequence to mora-based."""
    for (labels, moras, word_boundaries), expected in [
        (
            # Case: the IP boundary annotated at middle
            (
                [[0, 0, 1, 0, 0, 0, 0, 0, 0, 0]],
                [
                    [
                        "ワ",
                        "タ",
                        "シ",
                        "タ",
                        "チ",
                        "ガ",
                        "ム",
                        "ス",
                        "メ",
                        "ダ",
                        "ト",
                        "オ",
                        "モ",
                        "ッ",
                        "テ",
                        "イ",
                        "タ",
                    ]
                ],
                [
                    np.array(
                        [0, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 1, 0, 0, 1, 1, 1],
                        dtype=np.uint8,
                    ),
                ],
            ),
            [[1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]],
        ),
        (
            # Case: IP boundary annotated at last
            (
                [[0, 0, 1, 0, 0, 0, 0, 0, 0, 1]],
                [
                    [
                        "ワ",
                        "タ",
                        "シ",
                        "タ",
                        "チ",
                        "ガ",
                        "ム",
                        "ス",
                        "メ",
                        "ダ",
                        "ト",
                        "オ",
                        "モ",
                        "ッ",
                        "テ",
                        "イ",
                        "タ",
                    ],
                ],
                [
                    np.array(
                        [0, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 1, 0, 0, 1, 1, 1],
                        dtype=np.uint8,
                    ),
                ],
            ),
            [[1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]],
        ),
    ]:
        mora_labels = expand_word_label_to_mora(
            labels, moras, word_boundaries, "intonation_phrase_boundary"
        )
        assert expected == mora_labels


def test_accent_conversion():
    """Test function to convert ap-based accent sequence to mora-based."""
    for (
        ap_accents,
        accent_phrase_boundaries,
        accent_represent_mode,
        moras,
    ), expected in [
        # For binary
        # Correct label
        (
            (
                # NOTE: AP-based accent label requires remove padding index (=0)
                # e,g, torch.tensor([4]) means 3rd mora has accent nucleus
                torch.tensor([4]),
                # NOTE: accent phrase boundary label requires remove padding index (=0)
                # e,g, torch.tensor([1, 2, 1]) means 2nd mora has accent phrase boundary
                torch.tensor([1, 1, 1]),
                "binary",
                ["テ", "ス", "ト"],
            ),
            np.array([0, 0, 1]),
        ),
        # Wrong label: Out of mora
        (
            (
                torch.tensor([5]),
                torch.tensor([1, 1, 1]),
                "binary",
                ["テ", "ス", "ト"],
            ),
            np.array([0, 0, 0]),
        ),
        # Wrong label: wrong locattion (long-vowel)
        (
            (
                torch.tensor([5]),
                torch.tensor([1, 1, 1, 1]),
                "binary",
                ["テ", "ス", "ト", "ー"],
            ),
            np.array([0, 0, 0, 0]),
        ),
        # Wrong label: wrong locattion (sokuon)
        (
            (
                torch.tensor([5]),
                torch.tensor([1, 1, 1, 1]),
                "binary",
                ["テ", "ス", "ト", "ッ"],
            ),
            np.array([0, 0, 0, 0]),
        ),
        # For high_low
        # Correct label
        (
            (
                torch.tensor([4]),
                torch.tensor([1, 1, 1]),
                "high_low",
                ["テ", "ス", "ト"],
            ),
            np.array([0, 1, 1]),
        ),
        # Wrong label: Out of mora
        (
            (
                torch.tensor([5]),
                torch.tensor([1, 1, 1]),
                "high_low",
                ["テ", "ス", "ト"],
            ),
            np.array([0, 1, 1]),
        ),
        # Wrong label: wrong locattion (long-vowel)
        (
            (
                torch.tensor([5]),
                torch.tensor([1, 1, 1, 1]),
                "high_low",
                ["テ", "ス", "ト", "ー"],
            ),
            np.array([0, 1, 1, 1]),
        ),
        # Wrong label: wrong locattion (sokuon)
        (
            (
                torch.tensor([5]),
                torch.tensor([1, 1, 1, 1]),
                "high_low",
                ["テ", "ス", "ト", "ッ"],
            ),
            np.array([0, 1, 1, 1]),
        ),
    ]:
        mora_accents = _convert_ap_based_accent_to_mora_based_accent(
            ap_accents,
            accent_phrase_boundaries,
            accent_represent_mode,
            moras,
        )

        assert np.all(mora_accents == expected)


def test_multiple_task_score_calculation(test_log_sample):
    tasks = ["intonation_phrase_boundary", "accent_phrase_boundary", "accent_status"]
    expected_scores = {
        "intonation_phrase_boundary+accent_phrase_boundary+accent_status": 0.5203101920236337,
        "intonation_phrase_boundary+accent_phrase_boundary": 0.6410635155096012,
        "accent_phrase_boundary+accent_status": 0.6790989660265879,
    }

    scores = _calculate_multiple_task_scores(tasks, test_log_sample)

    # verify multiple task key is correct
    assert scores.keys() == expected_scores.keys()

    # verify scoring is currect
    for task_key in expected_scores.keys():
        assert_almost_equal(
            scores[task_key]["sentence_level_accuracy"],
            expected_scores[task_key],
        )


def test_init_model(default_config, default_vocab_path):
    """Verify init_model() works with default config."""
    feature_set = FeatureSet(default_vocab_path)

    for task_group, is_train in [
        (
            ["intonation_phrase_boundary", "accent_phrase_boundary", "accent_status"],
            True,
        ),
        (
            ["accent_phrase_boundary", "accent_status"],
            True,
        ),
        (
            ["accent_status"],
            True,
        ),
        (
            ["intonation_phrase_boundary", "accent_phrase_boundary", "accent_status"],
            False,
        ),
        (
            ["accent_phrase_boundary", "accent_status"],
            False,
        ),
        (
            ["accent_status"],
            False,
        ),
    ]:
        output = init_model(
            task_group, default_config, feature_set, device="cpu", is_train=is_train
        )

        if is_train:
            model, _, _, _ = output
        else:
            model = output

        print(model)


def test_is_adjective():
    for pos_tag, expect in [
        ("助動詞:*:*:*", False),
        ("動詞:一般:*:*", False),
        ("名詞:助動詞語幹:*:*", False),
        ("形容詞:一般:*:*", True),
        ("形状詞:タリ:*:*", False),
    ]:
        assert is_adjective(pos_tag) == expect


def test_is_noun():
    for pos_tag, expect in [
        ("助動詞:*:*:*", False),
        ("動詞:一般:*:*", False),
        ("名詞:助動詞語幹:*:*", True),
        ("形容詞:一般:*:*", False),
        ("形状詞:タリ:*:*", False),
    ]:
        assert is_noun(pos_tag) == expect


def test_is_verb():
    for pos_tag, expect in [
        ("助動詞:*:*:*", False),
        ("動詞:一般:*:*", True),
        ("名詞:助動詞語幹:*:*", False),
        ("形容詞:一般:*:*", False),
        ("形状詞:タリ:*:*", False),
    ]:
        assert is_verb(pos_tag) == expect


def test_parse_accent_con_type():
    for a_con_type, pos_tag, expect in [
        ("動詞%F2@0,形容詞%F2@-1,名詞%F1", "助動詞:*:*:*", "[UNK]"),
        ("動詞%F2@0,形容詞%F2@-1,名詞%F1", "動詞:一般:*:*", "F2"),
        ("動詞%F2@0,形容詞%F2@-1,名詞%F1", "名詞:助動詞語幹:*:*", "F1"),
        ("動詞%F2@0,形容詞%F2@-1,名詞%F1", "形容詞:一般:*:*", "F2"),
        ("動詞%F2@0,形容詞%F2@-1,名詞%F1", "形状詞:タリ:*:*", "[UNK]"),
    ]:
        assert parse_accent_con_type(a_con_type, pos_tag) == expect


def test_convert_open_jtalk_format_label():
    for labels, morph_boundary, expect in [
        # Fully aligned phrase boundary and word boundary
        (
            {
                "mora": [
                    [
                        "ナ",
                        "ダ",
                        "レ",
                        "デ",
                        "ド",
                        "ー",
                        "ロ",
                        "ガ",
                        "フ",
                        "サ",
                        "ガ",
                        "ッ",
                        "タ",
                        ".",
                    ],
                ],
                "intonation_phrase_boundary": [
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
                ],
                "accent_phrase_boundary": [[0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0]],
                "accent_status": [[0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0]],
            },
            [np.array([0, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 0, 1, 1], dtype=np.uint8)],
            {
                "accent_status": [0, 0, 1, 0, 3, 0, 0],
                "accent_phrase_boundary": [-1, 1, 0, 1, 0, 1, 1],
            },
        ),
        # Not aligned phrase boundary and word boundary (boundary in word)
        (
            {
                "mora": [
                    [
                        "ナ",
                        "ダ",
                        "レ",
                        "デ",
                        "ド",
                        "ー",
                        "ロ",
                        "ガ",
                        "フ",
                        "サ",
                        "ガ",
                        "ッ",
                        "タ",
                        ".",
                    ],
                ],
                "intonation_phrase_boundary": [
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
                ],
                "accent_phrase_boundary": [[0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0]],
                "accent_status": [[0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0]],
            },
            [np.array([0, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 0, 1, 1], dtype=np.uint8)],
            {
                "accent_status": [5, 0, 0, 0, 0, 0, 0],
                "accent_phrase_boundary": [-1, 1, 1, 1, 0, 1, 1],
            },
        ),
    ]:
        result = convert_open_jtalk_format_label(labels, morph_boundary)
        assert result == expect



# Copyright 2019 Allen Institute for AI
#  Apache 2.0  (http://www.apache.org/licenses/LICENSE-2.0)

"""
Conditional random field
"""

import math
from logging import getLogger
from typing import List, Optional, Tuple, Union

import torch

logger = getLogger(__name__)

VITERBI_DECODING = Tuple[List[int], float]  # a list of tags, and a viterbi score


def logsumexp(
    tensor: torch.Tensor, dim: int = -1, keepdim: bool = False
) -> torch.Tensor:
    """
    A numerically stable computation of logsumexp. This is mathematically equivalent to
    `tensor.exp().sum(dim, keep=keepdim).log()`.
    This function is typically used for summing log probabilities.
    # Parameters
    tensor : `torch.FloatTensor`, required.
        A tensor of arbitrary size.
    dim : `int`, optional (default = `-1`)
        The dimension of the tensor to apply the logsumexp to.
    keepdim: `bool`, optional (default = `False`)
        Whether to retain a dimension of size one at the dimension we reduce over.
    """
    max_score, _ = tensor.max(dim, keepdim=keepdim)
    if keepdim:
        stable_vec = tensor - max_score
    else:
        stable_vec = tensor - max_score.unsqueeze(dim)
    return max_score + (stable_vec.exp().sum(dim, keepdim=keepdim)).log()


def viterbi_decode(
    tag_sequence: torch.Tensor,
    transition_matrix: torch.Tensor,
    tag_observations: Optional[List[int]] = None,
    allowed_start_transitions: torch.Tensor = None,
    allowed_end_transitions: torch.Tensor = None,
    top_k: int = None,
):
    """
    Perform Viterbi decoding in log space over a sequence given a transition matrix
    specifying pairwise (transition) potentials between tags and a matrix of shape
    (sequence_length, num_tags) specifying unary potentials for possible tags per
    timestep.
    # Parameters
    tag_sequence : `torch.Tensor`, required.
        A tensor of shape (sequence_length, num_tags) representing scores for
        a set of tags over a given sequence.
    transition_matrix : `torch.Tensor`, required.
        A tensor of shape (num_tags, num_tags) representing the binary potentials
        for transitioning between a given pair of tags.
    tag_observations : `Optional[List[int]]`, optional, (default = `None`)
        A list of length `sequence_length` containing the class ids of observed
        elements in the sequence, with unobserved elements being set to -1. Note that
        it is possible to provide evidence which results in degenerate labelings if
        the sequences of tags you provide as evidence cannot transition between each
        other, or those transitions are extremely unlikely. In this situation we log a
        warning, but the responsibility for providing self-consistent evidence ultimately
        lies with the user.
    allowed_start_transitions : `torch.Tensor`, optional, (default = `None`)
        An optional tensor of shape (num_tags,) describing which tags the START token
        may transition *to*. If provided, additional transition constraints will be used for
        determining the start element of the sequence.
    allowed_end_transitions : `torch.Tensor`, optional, (default = `None`)
        An optional tensor of shape (num_tags,) describing which tags may transition *to* the
        end tag. If provided, additional transition constraints will be used for determining
        the end element of the sequence.
    top_k : `int`, optional, (default = `None`)
        Optional integer specifying how many of the top paths to return. For top_k>=1, returns
        a tuple of two lists: top_k_paths, top_k_scores, For top_k==None, returns a flattened
        tuple with just the top path and its score (not in lists, for backwards compatibility).
    # Returns
    viterbi_path : `List[int]`
        The tag indices of the maximum likelihood tag sequence.
    viterbi_score : `torch.Tensor`
        The score of the viterbi path.
    """
    if top_k is None:
        top_k = 1
        flatten_output = True
    elif top_k >= 1:
        flatten_output = False
    else:
        raise ValueError(
            f"top_k must be either None or an integer >=1. Instead received {top_k}"
        )

    sequence_length, num_tags = list(tag_sequence.size())

    has_start_end_restrictions = (
        allowed_end_transitions is not None or allowed_start_transitions is not None
    )

    if has_start_end_restrictions:

        if allowed_end_transitions is None:
            allowed_end_transitions = torch.zeros(num_tags)
        if allowed_start_transitions is None:
            allowed_start_transitions = torch.zeros(num_tags)

        num_tags = num_tags + 2
        new_transition_matrix = torch.zeros(num_tags, num_tags)
        new_transition_matrix[:-2, :-2] = transition_matrix

        # Start and end transitions are fully defined, but cannot transition between each other.

        allowed_start_transitions = torch.cat(
            [allowed_start_transitions, torch.tensor([-math.inf, -math.inf])]
        )
        allowed_end_transitions = torch.cat(
            [allowed_end_transitions, torch.tensor([-math.inf, -math.inf])]
        )

        # First define how we may transition FROM the start and end tags.
        new_transition_matrix[-2, :] = allowed_start_transitions
        # We cannot transition from the end tag to any tag.
        new_transition_matrix[-1, :] = -math.inf

        new_transition_matrix[:, -1] = allowed_end_transitions
        # We cannot transition to the start tag from any tag.
        new_transition_matrix[:, -2] = -math.inf

        transition_matrix = new_transition_matrix

    if tag_observations:
        if len(tag_observations) != sequence_length:
            raise ValueError(
                "Observations were provided, but they were not the same length "
                "as the sequence. Found sequence of length: {} and evidence: {}".format(
                    sequence_length, tag_observations
                )
            )
    else:
        tag_observations = [-1 for _ in range(sequence_length)]

    if has_start_end_restrictions:
        tag_observations = [num_tags - 2] + tag_observations + [num_tags - 1]
        zero_sentinel = torch.zeros(1, num_tags)
        extra_tags_sentinel = torch.ones(sequence_length, 2) * -math.inf
        tag_sequence = torch.cat([tag_sequence, extra_tags_sentinel], -1)
        tag_sequence = torch.cat([zero_sentinel, tag_sequence, zero_sentinel], 0)
        sequence_length = tag_sequence.size(0)

    path_scores = []
    path_indices = []

    if tag_observations[0] != -1:
        one_hot = torch.zeros(num_tags)
        one_hot[tag_observations[0]] = 100000.0
        path_scores.append(one_hot.unsqueeze(0))
    else:
        path_scores.append(tag_sequence[0, :].unsqueeze(0))

    # Evaluate the scores for all possible paths.
    for timestep in range(1, sequence_length):
        # Add pairwise potentials to current scores.
        summed_potentials = path_scores[timestep - 1].unsqueeze(2) + transition_matrix
        summed_potentials = summed_potentials.view(-1, num_tags)

        # Best pairwise potential path score from the previous timestep.
        max_k = min(summed_potentials.size()[0], top_k)
        scores, paths = torch.topk(summed_potentials, k=max_k, dim=0)

        # If we have an observation for this timestep, use it
        # instead of the distribution over tags.
        observation = tag_observations[timestep]
        # Warn the user if they have passed
        # invalid/extremely unlikely evidence.
        if tag_observations[timestep - 1] != -1 and observation != -1:
            if transition_matrix[tag_observations[timestep - 1], observation] < -10000:
                logger.warning(
                    "The pairwise potential between tags you have passed as "
                    "observations is extremely unlikely. Double check your evidence "
                    "or transition potentials!"
                )
        if observation != -1:
            one_hot = torch.zeros(num_tags)
            one_hot[observation] = 100000.0
            path_scores.append(one_hot.unsqueeze(0))
        else:
            path_scores.append(tag_sequence[timestep, :] + scores)
        path_indices.append(paths.squeeze())

    # Construct the most likely sequence backwards.
    path_scores_v = path_scores[-1].view(-1)
    max_k = min(path_scores_v.size()[0], top_k)
    viterbi_scores, best_paths = torch.topk(path_scores_v, k=max_k, dim=0)
    viterbi_paths = []
    for i in range(max_k):
        viterbi_path = [best_paths[i]]
        for backward_timestep in reversed(path_indices):
            viterbi_path.append(int(backward_timestep.view(-1)[viterbi_path[-1]]))
        # Reverse the backward path.
        viterbi_path.reverse()

        if has_start_end_restrictions:
            viterbi_path = viterbi_path[1:-1]

        # Viterbi paths uses (num_tags * n_permutations) nodes; therefore, we need to modulo.
        viterbi_path = [j % num_tags for j in viterbi_path]
        viterbi_paths.append(viterbi_path)

    if flatten_output:
        return viterbi_paths[0], viterbi_scores[0]

    return viterbi_paths, viterbi_scores


class ConditionalRandomField(torch.nn.Module):
    """
    This module uses the "forward-backward" algorithm to compute
    the log-likelihood of its inputs assuming a conditional random field model.
    See, e.g. http://www.cs.columbia.edu/~mcollins/fb.pdf
    # Parameters
    num_tags : `int`, required
        The number of tags.
    constraints : `List[Tuple[int, int]]`, optional (default = `None`)
        An optional list of allowed transitions (from_tag_id, to_tag_id).
        These are applied to `viterbi_tags()` but do not affect `forward()`.
        These should be derived from `allowed_transitions` so that the
        start and end transitions are handled correctly for your tag type.
    include_start_end_transitions : `bool`, optional (default = `True`)
        Whether to include the start and end transition parameters.
    """

    def __init__(
        self,
        num_tags: int,
        constraints: List[Tuple[int, int]] = None,
        include_start_end_transitions: bool = True,
    ) -> None:
        super().__init__()
        self.num_tags = num_tags

        # transitions[i, j] is the logit for transitioning from state i to state j.
        self.transitions = torch.nn.Parameter(torch.empty(num_tags, num_tags))

        # _constraint_mask indicates valid transitions (based on supplied constraints).
        # Include special start of sequence (num_tags + 1)
        # and end of sequence tags (num_tags + 2)
        if constraints is None:
            # All transitions are valid.
            constraint_mask = torch.full((num_tags + 2, num_tags + 2), 1.0)
        else:
            constraint_mask = torch.full((num_tags + 2, num_tags + 2), 0.0)
            for i, j in constraints:
                constraint_mask[i, j] = 1.0

        self._constraint_mask = torch.nn.Parameter(constraint_mask, requires_grad=False)

        # Also need logits for transitioning from "start" state and to "end" state.
        self.include_start_end_transitions = include_start_end_transitions
        if include_start_end_transitions:
            self.start_transitions = torch.nn.Parameter(torch.Tensor(num_tags))
            self.end_transitions = torch.nn.Parameter(torch.Tensor(num_tags))

        self.reset_parameters()

    def reset_parameters(self):
        torch.nn.init.xavier_normal_(self.transitions)
        if self.include_start_end_transitions:
            torch.nn.init.normal_(self.start_transitions)
            torch.nn.init.normal_(self.end_transitions)

    def _input_likelihood(
        self, logits: torch.Tensor, transitions: torch.Tensor, mask: torch.BoolTensor
    ) -> torch.Tensor:
        """
        Computes the (batch_size,) denominator term $Z(x)$, per example, for the log-likelihood
        This is the sum of the likelihoods across all possible state sequences.
        Args:
            logits (torch.Tensor): a (batch_size, sequence_length num_tags) tensor of
                unnormalized log-probabilities
            transitions (torch.Tensor): a (batch_size, num_tags, num_tags) tensor of
                transition scores
            mask (torch.BoolTensor): a (batch_size, sequence_length) tensor of masking flags
        Returns:
            torch.Tensor: (batch_size,) denominator term $Z(x)$, per example,
                for the log-likelihood
        """
        batch_size, sequence_length, num_tags = logits.size()

        # Transpose batch size and sequence dimensions
        mask = mask.transpose(0, 1).contiguous()
        logits = logits.transpose(0, 1).contiguous()

        # Initial alpha is the (batch_size, num_tags) tensor of likelihoods combining the
        # transitions to the initial states and the logits for the first timestep.
        if self.include_start_end_transitions:
            alpha = self.start_transitions.view(1, num_tags) + logits[0]
        else:
            alpha = logits[0]

        # For each i we compute logits for the transitions from timestep i-1 to timestep i.
        # We do so in a (batch_size, num_tags, num_tags) tensor where the axes are
        # (instance, current_tag, next_tag)
        for i in range(1, sequence_length):
            # The emit scores are for time i ("next_tag")
            # so we broadcast along the current_tag axis.
            emit_scores = logits[i].view(batch_size, 1, num_tags)
            # Transition scores are (current_tag, next_tag)
            # so we broadcast along the instance axis.
            transition_scores = transitions.view(1, num_tags, num_tags)
            # Alpha is for the current_tag, so we broadcast along the next_tag axis.
            broadcast_alpha = alpha.view(batch_size, num_tags, 1)

            # Add all the scores together and logexp over the current_tag axis.
            inner = broadcast_alpha + emit_scores + transition_scores

            # In valid positions (mask == True)
            # we want to take the logsumexp over the current_tag dimension
            # of `inner`. Otherwise (mask == False) we want to retain the previous alpha.
            alpha = logsumexp(inner, 1) * mask[i].view(batch_size, 1) + alpha * (
                ~mask[i]
            ).view(batch_size, 1)

        # Every sequence needs to end with a transition to the stop_tag.
        if self.include_start_end_transitions:
            stops = alpha + self.end_transitions.view(1, num_tags)
        else:
            stops = alpha

        # Finally we log_sum_exp along the num_tags dim, result is (batch_size,)
        return logsumexp(stops)

    def _joint_likelihood(
        self,
        logits: torch.Tensor,
        transitions: torch.Tensor,
        tags: torch.Tensor,
        mask: torch.BoolTensor,
    ) -> torch.Tensor:
        """Computes the numerator term for the log-likelihood, which is just score(inputs, tags)
        Args:
            logits (torch.Tensor): a (batch_size, sequence_length num_tags) tensor
                of unnormalized log-probabilities
            transitions (torch.Tensor): a (batch_size, num_tags, num_tags) tensor
                of transition scores
            tags (torch.Tensor): output tag sequences (batch_size, sequence_length) $y$
                for each input sequence
            mask (torch.BoolTensor): a (batch_size, sequence_length) tensor of masking flags
        Returns:
            torch.Tensor: numerator term for the log-likelihood,
                which is just score(inputs, tags)
        """
        batch_size, sequence_length, _ = logits.data.shape

        # Transpose batch size and sequence dimensions:
        logits = logits.transpose(0, 1).contiguous()
        mask = mask.transpose(0, 1).contiguous()
        tags = tags.transpose(0, 1).contiguous()

        # Start with the transition scores from start_tag to the first tag in each input
        if self.include_start_end_transitions:
            score = self.start_transitions.index_select(0, tags[0])
        else:
            score = 0.0

        # Add up the scores for the observed transitions and all the inputs but the last
        for i in range(sequence_length - 1):
            # Each is shape (batch_size,)
            current_tag, next_tag = tags[i], tags[i + 1]

            # The scores for transitioning from current_tag to next_tag)
            transition_score = transitions[current_tag.view(-1), next_tag.view(-1)]

            # The score for using current_tag
            emit_score = logits[i].gather(1, current_tag.view(batch_size, 1)).squeeze(1)

            # Include transition score if next element is unmasked,
            # input_score if this element is unmasked.
            score = score + transition_score * mask[i + 1] + emit_score * mask[i]

        # Transition from last state to "stop" state.
        # To start with, we need to find the last tag
        # for each instance.
        last_tag_index = mask.sum(0).long() - 1
        last_tags = tags.gather(0, last_tag_index.view(1, batch_size)).squeeze(0)

        # Compute score of transitioning to `stop_tag` from each "last tag".
        if self.include_start_end_transitions:
            last_transition_score = self.end_transitions.index_select(0, last_tags)
        else:
            last_transition_score = 0.0

        # Add the last input if it's not masked.
        last_inputs = logits[-1]  # (batch_size, num_tags)
        last_input_score = last_inputs.gather(
            1, last_tags.view(-1, 1)
        )  # (batch_size, 1)
        last_input_score = last_input_score.squeeze()  # (batch_size,)

        score = score + last_transition_score + last_input_score * mask[-1]

        return score

    def forward(
        self, inputs: torch.Tensor, tags: torch.Tensor, mask: torch.BoolTensor = None
    ) -> torch.Tensor:
        """Computes the log likelihood for the given batch of input sequences $(x,y)$
        Args:
            inputs (torch.Tensor): (batch_size, sequence_length, num_tags) tensor
                of logits for the inputs $x$
            tags (torch.Tensor): (batch_size, sequence_length) tensor of tags $y$
            mask (torch.BoolTensor, optional): (batch_size, sequence_length) tensor
                of masking flags. Defaults to None.
        Returns:
            torch.Tensor: (batch_size,) log likelihoods $log P(y|x)$ for each input
        """
        if mask is None:
            mask = torch.ones(*tags.size(), dtype=torch.bool, device=inputs.device)
        else:
            # The code below fails in weird ways if this isn't a bool tensor, so we make sure.
            mask = mask.to(torch.bool)

        log_denominator = self._input_likelihood(inputs, self.transitions, mask)
        log_numerator = self._joint_likelihood(inputs, self.transitions, tags, mask)

        return torch.sum(log_numerator - log_denominator)

    def viterbi_tags(
        self, logits: torch.Tensor, mask: torch.BoolTensor = None, top_k: int = None
    ) -> Union[List[VITERBI_DECODING], List[List[VITERBI_DECODING]]]:
        """
        Uses viterbi algorithm to find most likely tags for the given inputs.
        If constraints are applied, disallows all other transitions.
        Returns a list of results, of the same size as the batch (one result per batch member)
        Each result is a List of length top_k, containing the top K viterbi decodings
        Each decoding is a tuple  (tag_sequence, viterbi_score)
        For backwards compatibility, if top_k is None, then instead returns a flat list of
        tag sequences (the top tag sequence for each batch item).
        """
        if mask is None:
            mask = torch.ones(*logits.shape[:2], dtype=torch.bool, device=logits.device)

        if top_k is None:
            top_k = 1
            flatten_output = True
        else:
            flatten_output = False

        _, max_seq_length, num_tags = logits.size()

        # Get the tensors out of the variables
        logits, mask = logits.data, mask.data

        # Augment transitions matrix with start and end transitions
        start_tag = num_tags
        end_tag = num_tags + 1
        transitions = torch.full(
            (num_tags + 2, num_tags + 2), -10000.0, device=logits.device
        )

        # Apply transition constraints
        constrained_transitions = self.transitions * self._constraint_mask[
            :num_tags, :num_tags
        ] + -10000.0 * (1 - self._constraint_mask[:num_tags, :num_tags])
        transitions[:num_tags, :num_tags] = constrained_transitions.data

        if self.include_start_end_transitions:
            transitions[
                start_tag, :num_tags
            ] = self.start_transitions.detach() * self._constraint_mask[
                start_tag, :num_tags
            ].data + -10000.0 * (
                1 - self._constraint_mask[start_tag, :num_tags].detach()
            )
            transitions[
                :num_tags, end_tag
            ] = self.end_transitions.detach() * self._constraint_mask[
                :num_tags, end_tag
            ].data + -10000.0 * (
                1 - self._constraint_mask[:num_tags, end_tag].detach()
            )
        else:
            transitions[start_tag, :num_tags] = -10000.0 * (
                1 - self._constraint_mask[start_tag, :num_tags].detach()
            )
            transitions[:num_tags, end_tag] = -10000.0 * (
                1 - self._constraint_mask[:num_tags, end_tag].detach()
            )

        best_paths = []
        # Pad the max sequence length by 2 to account for start_tag + end_tag.
        tag_sequence = torch.empty(
            max_seq_length + 2, num_tags + 2, device=logits.device
        )

        for prediction, prediction_mask in zip(logits, mask):
            mask_indices = prediction_mask.nonzero(as_tuple=False).squeeze()
            masked_prediction = torch.index_select(prediction, 0, mask_indices)
            sequence_length = masked_prediction.shape[0]

            # Start with everything totally unlikely
            tag_sequence.fill_(-10000.0)
            # At timestep 0 we must have the START_TAG
            tag_sequence[0, start_tag] = 0.0
            # At steps 1, ..., sequence_length we just use the incoming prediction
            tag_sequence[1 : (sequence_length + 1), :num_tags] = masked_prediction
            # And at the last timestep we must have the END_TAG
            tag_sequence[sequence_length + 1, end_tag] = 0.0

            # We pass the tags and the transitions to `viterbi_decode`.
            viterbi_paths, viterbi_scores = viterbi_decode(
                tag_sequence=tag_sequence[: (sequence_length + 2)],
                transition_matrix=transitions,
                top_k=top_k,
            )
            top_k_paths = []
            for viterbi_path, viterbi_score in zip(viterbi_paths, viterbi_scores):
                # Get rid of START and END sentinels and append.
                viterbi_path = viterbi_path[1:-1]
                top_k_paths.append((viterbi_path, viterbi_score.item()))
            best_paths.append(top_k_paths)

        if flatten_output:
            return [top_k_paths[0] for top_k_paths in best_paths]

        return best_paths
